"""
This module regroups the classes describing
the requests made by an agent during the simulation.

Requests are used for the interaction between users and operators.

Stops describe the action of a service vehicle, for instance picking up a user.
"""


class Request:
    """
    Class describing a request made by an agent.
    """

    GET_REQUEST = "GET"
    PUT_REQUEST = "PUT"
    TAXI_REQUEST = "TAXI"

    STR_LIST = ["structure", "position", "type"]

    def __init__(
        self, agent, timestamp, structure=None, position=None, request_type=None, prebooked=None
    ):
        """
        Creation of a new request.

        :param agent: agent making the request
        :param timestamp: time when the request was made
        :param structure: structure requested by the agent
        :param position: request's position
        :param request_type: type of request
        """

        # requesting agent
        self.agent = agent

        # timestamp of the request
        self.timestamp = timestamp

        # requested structure
        self.structure = structure

        # target position of the request
        self.position = position

        # request type
        self.type = request_type

        # prebooked request
        self.prebooked = prebooked

        # SimPy Event, can be used by the agent as a callback
        self.event_ = None

        # request success
        self.success = None

        # request result
        self.result = None

        # sequence of waiting times due to request
        self.waitSequence = []

    def set_request_event(self, event=None):
        """
        Set the event attribute with the given SimPy event.

        By default, the a basic Event object is added, which can be triggered
        using the succeed method.

        If the request has no agent, try to get the scheduler from the structure.

        :param event: SimPy Event object
        """

        if event is None:
            if self.agent is None:
                event = self.structure.sim.scheduler.new_event_object()
            else:
                event = self.agent.sim.scheduler.new_event_object()

        self.event_ = event

    def succeed(self, new_event=None):
        """
        Trigger the request event as successful and replace it with the new event.

        :param new_event: SimPy Event object
        """

        former_event = self.event_

        self.set_request_event(new_event)

        former_event.succeed()

    def fail(self, new_event=None):
        """
        Trigger the request event as failed and replace it with the new event.

        :param new_event: SimPy Event object
        """
        former_event = self.event_

        self.set_request_event(new_event)

        # TODO : need an exception argument ?
        former_event.fail()

    def cancel(self):
        """
        Set success to false and call the structure cancel method.
        """

        self.success = False

        self.structure.cancel_request(self)

    def __str__(self):
        """
        Give string display to the request.

        :return: string with some of the request information
        """

        fill_string = "requestTime=" + str(self.timestamp) + ", agent=" + str(self.agent.id)

        for attr_name in self.STR_LIST:
            attribute = self.__getattribute__(attr_name)
            if attribute is not None:
                if attr_name == "structure":
                    attribute = attribute.id
                fill_string += ", " + attr_name + "=" + str(attribute)

        fill_string += ", success={}, result={}, waitDurations={}"

        return fill_string.format(self.success, self.result, self.waitSequence)

    def __repr__(self):
        return self.__str__()


class StationRequest(Request):
    """
    Class describing a request made to a vehicle station.
    """

    def __init__(self, agent, timestamp, station, request_type):
        """
        Creation of a new station request

        :param agent: agent making the request
        :param timestamp: time when the request was made
        :param station: station requested by the agent
        :param request_type: type of the request
        """

        super().__init__(agent, timestamp, structure=station, request_type=request_type)


class TripRequest(Request):
    """
    Class describing a trip request made to a service operator.
    """

    STR_LIST = ["structure", "id", "position", "type"]

    def __init__(self, agent, timestamp, operator, number, request_id, trip_id=None):
        """
        Creation of a new trip request.

        :param agent: agent making the request
        :param timestamp: time when the request was made
        :param operator: operator requested by the agent
        :param number: number of seats requested
        :param request_id: id of the trip request, stored by the operator
        :param trip_id: if of the request trip, or None
        """

        super().__init__(agent, timestamp, structure=operator)

        # trip request id
        self.id = request_id

        # store the id of the request trip, if there is one
        self.tripId = trip_id

        # store the agent number
        self.number = number

        # store the direct travel time
        self.directTravelTime = None

        # trip stops

        # pickup stop
        self.pickup = None

        # dropoff stop
        self.dropoff = None

        # stops events

        # pickup event
        self.pickupEvent_ = self.structure.sim.scheduler.new_event_object()

        # dropoff event
        self.dropoffEvent_ = self.structure.sim.scheduler.new_event_object()

    def set_stops(self, pickup_request, dropoff_request):
        """
        Set the pickup and dropoff attributes of the request.

        :param pickup_request: UserStop object with type "GET"
        :param dropoff_request: UserStop object with type "PUT"
        """

        # set agent number
        pickup_request.number = self.number
        dropoff_request.number = self.number

        # set pickup twin stop
        pickup_request.set_twin_stop(dropoff_request)

        # set pickup stop
        self.pickup = pickup_request

        # set dropoff twin stop
        dropoff_request.set_twin_stop(pickup_request)

        # set dropoff stop
        self.dropoff = dropoff_request

    def set_trip(self, trip_id):
        """
        Set the pickup and dropoff trip id, if they are UserStops.

        :param trip_id:
        """

        self.tripId = trip_id

        if isinstance(self.pickup, UserStop) and isinstance(self.dropoff, UserStop):
            self.pickup.tripId = trip_id
            self.dropoff.tripId = trip_id

    def pickup_succeed(self):
        self.pickupEvent_.succeed()

    def dropoff_succeed(self):
        self.dropoffEvent_.succeed()


class Stop:
    """
    Class describing a stop in a service vehicle planning
    """

    GET_REQUEST = "GET"
    PUT_REQUEST = "PUT"
    TAXI_REQUEST = "TAXI"
    STOP_POINT = "STOP_POINT"
    REPOSITIONING = "REPOSITIONING"

    def __init__(self, position, stop_type):
        """
        Creation of a new stop.

        As we make copies (deepcopy) of the stops for testing plannings,
        they cannot contain attributes such as Agent or SimPy.Event objects.

        :param position: position of the stop in the environment
        :param stop_type: type of stop, either "GET", "PUT" or "STOP_POINT"
        """

        # Indicates the way to process the stop
        self.type = stop_type

        # position of the stop in the environment
        self.position = position

        # planned or effective arrival time
        self.arrivalTime = None

        # planned or effective departure time
        self.departureTime = None

        # effective arrival time
        self.effectiveArrivalTime = None

        # effective departure time
        self.effectiveDepartureTime = None


class UserStop(Stop):
    """
    Class describing the stop of a user, either a pickup or dropoff.

    User stops can be used as is in the plannings, or grouped in stop points.

    User stops should be grouped by pair (pickup and dropoff) in TripRequest objects,
    using the setup_stops method.
    """

    def __init__(
        self,
        stop_type,
        position,
        request_id,
        requested_time=None,
        max_time=None,
        max_travel_time=None,
        stop_point_id=None,
        trip_id=None,
    ):
        """
        Creation of a new user stop.

        :param stop_type: type of user stop, either "GET" or "PUT"
        :param position: position of the user stop in the environment
        :param request_id: id of the corresponding TripRequest
        :param requested_time: time requested for the stop.
            Default is None, no min constraint on the stop service time.
        :param max_time: maximum service time accepted.
            Default is None, no max constraint on the stop service time.
        :param max_travel_time: maximum duration of the TripRequest travel.
            Default is None, no constraint on the trip travel time.
        :param stop_point_id: id of the stop point, if relevant
        :param trip_id: id of the requested trip, if relevant
        """

        super().__init__(position, stop_type)

        # id of the request made to the service operator
        self.requestId = request_id

        # twin user stop, the other stop of the same trip request
        self.twin = None

        # store the agent number
        self.number = None

        # id of the stop point if the user stop is part of a stop point
        self.stopPoint = stop_point_id

        # id of the requested trip, used to match the service vehicle. None matches all trips
        self.trip = trip_id

        # constraints on the service time

        # requested stop time (and then minimum stop time)
        # if None, the stop can be processed up as soon as possible
        self.requestedTime = requested_time

        # maximum process time
        # if None, the stop can be processed as late as wanted
        self.maxTime = max_time

        # maximum travel time
        # if None, no constraint on the trip travel time
        self.maxTravelTime = max_travel_time

    def get_process_time(self):
        """
        Get the time of the effective stop processing.

        Pickups are processed at departure time, and dropoffs
        are processed at arrival time.

        :return: process time of the stop
        """

        if self.type == self.GET_REQUEST:
            return self.departureTime
        elif self.type == self.PUT_REQUEST:
            return self.arrivalTime
        else:
            raise ValueError("Unsupported type {} for UserStop".format(self.type))

    def set_twin_stop(self, twin_stop):
        """
        Set the user stop twin stop.

        :param twin_stop: UserStop object
        """

        self.twin = twin_stop

    def is_feasible(self):
        """
        Check if the stop point is feasible, ie if the time constraints are respected.

        :return: Boolean indicating if the user stop is feasible
        """

        travel_time = None

        # compute the stop service time and travel time if relevant
        if self.type == self.GET_REQUEST:
            service_time = self.departureTime

            if service_time is not None and self.twin.arrivalTime is not None:
                travel_time = self.twin.arrivalTime - service_time

        elif self.type == self.PUT_REQUEST:
            service_time = self.arrivalTime

            if service_time is not None and self.twin.arrivalTime is not None:
                travel_time = service_time - self.twin.departureTime

        else:
            return False

        # check the min service time constraint
        if self.requestedTime is None:
            after_requested = True
        else:
            after_requested = self.requestedTime <= service_time

        # check the max service time constraint
        if self.maxTime is None:
            before_max = True
        else:
            before_max = service_time <= self.maxTime

        # check the max travel time constraint
        if self.maxTravelTime is None or travel_time is None:
            not_too_long = True
        else:
            not_too_long = travel_time <= self.maxTravelTime

        return after_requested and before_max and not_too_long

    def __str__(self):
        """
        Give a string display to the user stop.

        :return: string with some of the user stop information
        """

        fill_string = (
            "[requestId={}, stopType={}, position={}, arrivalTime={}, "
            "departureTime={}, requestedTime={}, maxTime={}]"
        )

        return fill_string.format(
            self.requestId,
            self.type,
            self.position,
            self.arrivalTime,
            self.departureTime,
            self.requestedTime,
            self.maxTime,
        )

    def __repr__(self):
        return self.__str__()


class StopPoint(Stop):
    """
    Class describing a stop point, where several user stops can be grouped.

    Stop points may be serviced by several vehicles following different trips.
    Servicing a stop point means servicing all the user stops that correspond
    to the service vehicle trip
    """

    def __init__(self, position, stop_id, name="unnamed_stop"):
        """
        Creation of a new stop point.

        :param position: position of the stop in the environment
        :param stop_id: id of the stop point, stored by the operator
        :param name: name of the stop point, for instance "République"
        """

        super().__init__(position, self.STOP_POINT)

        # id of the stop point, used for describing the service trips
        self.id = stop_id

        # name of the stop point, for a better readability
        self.name = name

        # list of user pickups at this stop
        self.pickupList = []

        # list of user dropoffs at this stop
        self.dropoffList = []

        # dict of the arrival times of this stop point, of the form {trip_id: arrival_time}
        self.arrivalTime = dict()

        # dict of the departure times of this stop point, of the form {trip_id: departure_time}
        self.departureTime = dict()

        # dict of the effective arrival times of this stop point, of the form {trip_id: arrival_time}
        self.effectiveArrivalTime = dict()

        # dict of the effective departure times of this stop point, of the form {trip_id: departure_time}
        self.effectiveDepartureTime = dict()

    def __str__(self):
        """
        Give a string display to the stop point.

        :return: string with some of the stop point information
        """
        return "[name={}, id={}, position={}]".format(self.name, self.id, self.position)

    def __repr__(self):
        return self.__str__()


class Operation(Stop):
    def __init__(self, stop_type, position, total, targets=None, station_id=None):
        """
        Creation of a new operation.

        :param stop_type: type of the operation (REPOSITIONING, ..)
        :param position: position in the environment
        :param total: relative number corresponding to operations realised
        :param targets: list of operation targets, or None
        :param station_id: id of the target station, or None
        """

        super().__init__(position, stop_type)

        # operations realised at this stop (may be negative, for instance when picking up vehicles)
        self.total = total

        # targets of the operation (None if there are no specific targets)
        self.targets = targets

        # id of the station, if there is one
        self.station = station_id

    def __str__(self):
        """
        Give a string display to the operation.

        :return: string with some of the operation information
        """

        return "[type={}, total={}, position={}, station={}]".format(
            self.type, self.total, self.position, self.station
        )

    def __repr__(self):
        return self.__str__()



"""
This module contains event classes that compose agents' traces
"""


class Event:
    """
    This class represents the events contained in the trace

    It should be extended to describe the different kinds of events of the agents
    """

    def __init__(self, time, message=""):
        """

        :param time: timestamp of the traced event
        :param message: eventual message to be added to the event
        """
        self.timestamp = time
        self.message = message

    def __str__(self):
        """
        Gives a string display to the event
        """

        if self.message == "":
            return "[{}, {}]: ".format(self.timestamp, self.__class__.__name__)
        else:
            return "[{}, {}, {}]: ".format(self.message, self.__class__.__name__, self.timestamp)


class InputEvent(Event):
    """
    This event describes the generation of a traced element.
    """

    def __init__(self, time, element, message=""):
        """
        Creates a generation event
        :param time: timestamp of the generation
        :param element: generated element
        :param message: eventual message to be added to the event
        """
        super().__init__(time, message=message)
        self.element = element

    def __str__(self):
        return super().__str__() + "generatedElement={}, type={}".format(
            self.element, type(self.element)
        )


class MoveEvent(Event):
    """
    This event describes an agent moving
    """

    def __init__(self, time, move_distance, move_duration, mode, message=""):
        """
        Creates a moving event
        :param time: timestamp of departure
        :param move_distance: move distance
        :param move_duration: time spent moving
        :param mode: transport mode ("walk", "bike", "drive",..)
        :param message: eventual message to be added to the event
        """

        super().__init__(time, message=message)

        self.distance = move_distance
        self.duration = move_duration
        self.mode = mode


class RouteEvent(MoveEvent):
    """
    This event describes the route of an agent
    """

    def __init__(self, time, route_data, mode, message=""):
        """
        Creates a route event.

        :param time: timestamp of the route departure
        :param route_data: {"route": list of nodes,
            "length": list of link lengths, "time": list of link durations}
        :param mode: transport mode ("walk", "bike", "drive",..)
        :param message: eventual message to be added to the event
        :return:
        """

        super().__init__(
            time, sum(route_data["length"]), sum(route_data["time"]), mode, message=message
        )

        self.data = route_data

    def set_route_data(self, route_data):
        self.data = route_data
        self.distance = sum(route_data["length"])
        self.duration = sum(route_data["time"])

    def __str__(self):
        return super().__str__() + "mode={}, start={}, end={}, duration={}, distance={}".format(
            self.mode, self.data["route"][0], self.data["route"][-1], self.duration, self.distance
        )


class PositionChangeEvent(MoveEvent):
    """
    This event describes a position change of an agent
    """

    def __init__(self, time, origin, destination, distance, duration, mode, message=""):
        """
        Creates a position change event
        :param destination:
        :param distance:
        :param duration:
        :param time: timestamp of the traced event
        :param origin: describes a position, e.g. a node id
        :param mode: transport mode used to reach this position
        :param message: eventual message to be added to the event
        """
        super().__init__(time, distance, duration, mode, message)

        self.origin = origin
        self.destination = destination

    def __str__(self):
        return super().__str__() + "mode={}, start={}, end={}, duration={}, distance={}".format(
            self.mode, self.origin, self.destination, self.duration, self.distance
        )


class WaitEvent(Event):
    """
    This event describes the a waiting agent
    """

    def __init__(self, time, reason, waiting_time, message=""):
        """
        Creates a wait event
        :param time: timestamp of the waiting start
        :param reason: reason of the waiting event
        :param waiting_time: waited duration
        :param message: eventual message to be added to the event
        """
        super().__init__(time, message)
        self.reason = reason
        self.waiting_time = waiting_time

    def __str__(self):
        return super().__str__() + "waitedTime={}, reason={}".format(self.waiting_time, self.reason)


class IdleEvent(Event):
    """
    This event describes an idle agent
    """

    def __init__(self, time, idle_duration, message=""):
        """
        Creates an idle event
        :param time: start of the idle period
        :param idle_duration: duration of the idle period
        :param message: eventual message to be added to the event
        """

        super().__init__(time, message)
        self.duration = idle_duration

    def __str__(self):
        return super().__str__() + "idleDuration={}".format(self.duration)


class ServiceEvent(Event):
    """
    This event describes a status change of an agent service.
    """

    def __init__(self, time, former_status, new_status, message=""):
        super().__init__(time, message)
        # former and new status values
        self.former = former_status
        self.new = new_status

    def __str__(self):
        return super().__str__() + "formerStatus={}, newStatus={}".format(self.former, self.new)


class RequestEvent(Event):
    """
    This event describes a user request
    """

    def __init__(self, time, request, message=""):
        """
        Creates a request event
        :param time: timestamp of the request
        :param request: Request object
        :param message: eventual message to be added to the event
        """
        super().__init__(time, message)
        self.request = request

    def __str__(self):
        return super().__str__() + str(self.request)


class StopEvent(Event):
    """
    This event describes the processing of a Stop
    """

    def __init__(self, time, operator, service_vehicle, trip, stop, message=""):
        super().__init__(time, message)

        self.operator = operator
        self.serviceVehicle = service_vehicle
        self.stop = stop
        self.trip = trip

        # list of requests served during dropoff
        self.dropoffs = []
        self.dropoff_time = None

        # list of requests served during pickup
        self.pickups = []
        self.pickup_time = None

    def set_dropoffs(self, dropoffs, dropoff_time):
        if dropoffs is None:
            dropoffs = []
        elif not isinstance(dropoffs, list):
            dropoffs = [dropoffs]

        self.dropoffs = dropoffs
        self.dropoff_time = dropoff_time

    def set_pickups(self, pickups, pickup_time):
        if pickups is None:
            pickups = []
        elif not isinstance(pickups, list):
            pickups = [pickups]

        self.pickups = pickups
        self.pickup_time = pickup_time

    def __str__(self):
        return super().__str__() + "stop={}, trip={}, serviceVehicle={}".format(
            self.stop, self.trip, self.serviceVehicle.id
        )


# deprecated, use StopEvent
# class PickupEvent(StopEvent):
#     """
#     This event describes the processing of a pickup
#     """
#
#     def __init__(self, time, operator, service_vehicle, trip, stop, pickups, message=""):
#
#         super().__init__(time, operator, service_vehicle, trip, stop, message=message)
#
#         if not isinstance(pickups, list):
#             pickups = [pickups]
#
#         self.pickups = pickups
#
#     def __str__(self):
#
#         return super().__str__() + "serviceVehicle={}, stop={}, trip={}, pickups={}"\
#             .format(self.serviceVehicle.id, str(self.stop), self.trip, self.pickups)
#
#
# # deprecated, use StopEvent
# class DropoffEvent(StopEvent):
#     """
#     This event describes the processing of a dropoff
#     """
#
#     def __init__(self, time, operator, service_vehicle, trip, stop, dropoffs, message=""):
#
#         super().__init__(time, operator, service_vehicle, trip, stop, message=message)
#
#         if not isinstance(dropoffs, list):
#             dropoffs = [dropoffs]
#
#         self.dropoffs = dropoffs
#
#     def __str__(self):
#
#         return super().__str__() + "serviceVehicle={}, stop={}, trip={}, dropoffs={}"\
#             .format(self.serviceVehicle.id, str(self.stop), self.trip, self.dropoffs)
#


class StaffOperationEvent(Event):
    """
    This event describes a staff operation
    """

    def __init__(self, time, staff, total, goal, targets=None, structure=None, message=""):
        """
        Create a staff operation event.

        :param time: timestamp of the staff operation event
        :param staff: staff realising the operation
        :param total: relative number corresponding to operations realised
        :param goal: objective of operation to realise
        :param targets: list of operation targets
        :param structure: structure where the operation is realised
        """

        super().__init__(time, message)

        self.staff = staff
        self.total = total
        self.goal = goal
        self.targets = targets
        if targets is None:
            self.targets = []
        self.structure = structure

    def __str__(self):
        if self.structure is None:
            struct = None
        else:
            struct = self.structure.id
        return super().__str__() + "staff={}, total={}, structure={}".format(
            self.staff.id, self.total, struct
        )


class GetVehicleEvent(Event):
    """
    This event describes an agent getting a vehicle
    """

    def __init__(self, time, agent, vehicle, message=""):
        """
        Creates a get vehicle event
        :param time: timestamp of the get event
        :param agent: new occupant
        :param vehicle: concerned vehicle
        :param message: eventual message to be added to the event
        """
        super().__init__(time, message)
        self.agent = agent
        self.vehicle = vehicle

    def __str__(self):
        return super().__str__() + "agent={}, getVehicle={}".format(self.agent.id, self.vehicle.id)


class LeaveVehicleEvent(Event):
    """
    This event describes an agent returning a vehicle
    """

    def __init__(self, time, agent, vehicle, message=""):
        """
        Creates a leave vehicle event
        :param time: timestamp of the return event
        :param agent: agent returning vehicle
        :param vehicle: concerned vehicle
        :param message: eventual message to be added to the event
        """
        super().__init__(time, message)
        self.agent = agent
        self.vehicle = vehicle

    def __str__(self):
        return super().__str__() + "agent={}, leaveVehicle={}".format(
            self.agent.id, self.vehicle.id
        )


class LeaveSystemEvent(Event):
    """
    This event describes an agent leaving the simulation
    """

    def __init__(self, time, message=""):
        super().__init__(time, message=message)

    def __str__(self):
        return super().__str__() + "leavingSystem={}".format(self.timestamp)


class DestinationReachedEvent(Event):
    """
    This event describes an agent reaching its destination
    """

    def __init__(self, time, message=""):
        super().__init__(time, message=message)

    def __str__(self):
        return super().__str__() + "arrivalTime={}".format(self.timestamp)


class LeaveSimulationEvent(Event):
    """
    This event describes an agent leaving the simulation.
    """

    def __init__(self, time, agent, cause, message=""):
        super().__init__(time, message=message)

        # agent leaving the simulation
        self.agent = agent

        #
        self.cause = cause

    def __str__(self):
        return super().__str__() + "agent={}, cause={}".format(self.agent, self.cause)


# class EndOfSimulationEvent(Event):
#     """
#     This event describes the end of the simulation
#     """
#
#     def __init__(self, time, message=""):
#         """
#         Creates an end of simulation event
#         :param time: timestamp of simulation end
#         :param message: eventual message to be added to the event
#         """
#         super().__init__(time, message=message)
#
#     def __str__(self):
#
#         return super().__str__() + "simulationEndTime={}" \
#             .format(self.timestamp)



from starling_sim.utils.utils import (
    new_point_feature,
    new_line_string_feature,
    new_multi_polygon_feature,
)
from starling_sim.basemodel.trace.events import InputEvent, RouteEvent, PositionChangeEvent

import logging


def create_point_feature(geojson_output, element, agent_id=None, icon_type=None, agent_type=None):
    point = get_element_point(geojson_output, element)
    feature = new_point_feature(point_localisation=point)
    add_agent_id(feature, element, agent_id)
    add_icon_type(feature, element, icon_type)
    add_agent_type(feature, element, agent_type)

    return feature


def create_line_string_feature(
    geojson_output, element, agent_id=None, icon_type=None, agent_type=None
):
    localisations, timestamps = get_element_line_string(geojson_output, element)
    feature = new_line_string_feature(localisations)
    feature["properties"]["timestamps"] = timestamps
    add_agent_id(feature, element, agent_id)
    add_icon_type(feature, element, icon_type)
    add_agent_type(feature, element, agent_type)

    return feature


def create_multi_polygon_feature(
    geojson_output, element, agent_id=None, icon_type=None, agent_type=None
):
    polygon_list = get_element_multi_polygon(geojson_output, element)
    feature = new_multi_polygon_feature(polygon_list)
    add_agent_id(feature, element, agent_id)
    add_icon_type(feature, element, icon_type)
    add_agent_type(feature, element, agent_type)

    return feature


def get_element_point(geojson_output, element):
    return geojson_output.sim.environment.get_localisation(element.position)[::-1]


def get_element_multi_polygon(geojson_output, element):
    zone_polygon = element.serviceZone.loc[0, "geometry"]
    linear_ring_coordinates = []
    for coord in zone_polygon.exterior.coords:
        linear_ring_coordinates.append(list(coord))
    polygon = [linear_ring_coordinates]
    polygon_list = [polygon]

    return polygon_list


def get_element_line_string(geojson_output, element):
    agent = element

    if not hasattr(agent, "origin"):
        return None

    # lists of localisations [lng, lat] and timestamps
    localisations = []
    timestamps = []

    # we add a position at 0 for the viz
    localisations.append(geojson_output.sim.environment.get_localisation(agent.origin))
    timestamps.append(0)

    # get agent's trace
    trace_list = agent.trace.eventList

    # start processing the trace
    i = 0

    while i < len(trace_list):
        # get the current event
        event = trace_list[i]

        # for the input event, add origin and input time
        if isinstance(event, InputEvent):
            # the input event should always come first
            if i == 0:
                localisations.append(geojson_output.sim.environment.get_localisation(agent.origin))
                current_time = event.timestamp
                timestamps.append(current_time)
            else:
                logging.warning("InputEvent does not come first in {}'s trace".format(agent.id))

        # for a route event, add all localisations
        # of the route (origin and dest included)
        if isinstance(event, RouteEvent):
            # get route mode
            mode = event.mode

            # get the list of route localisations and timestamps
            route_positions, route_timestamps = route_localisations(
                event, geojson_output.sim.scenario["limit"], geojson_output.graphs[mode]
            )

            # add it to the agent's lists
            localisations = localisations + route_positions
            timestamps = timestamps + route_timestamps

        # for a position change event, add the localisations and timestamps
        elif isinstance(event, PositionChangeEvent):
            # get the move mode
            mode = event.mode

            # add it to the agent's lists
            localisations.append(geojson_output.graphs[mode].position_localisation(event.origin))
            timestamps.append(event.timestamp)

            localisations.append(
                geojson_output.graphs[mode].position_localisation(event.destination)
            )
            timestamps.append(event.timestamp + event.duration)

        i += 1

    # add a lasting position for viz
    localisations.append(localisations[-1])
    timestamps.append(99999)

    # reverse lat and lon
    new_locs = []
    for loc in localisations:
        new_locs.append([loc[1], loc[0]])
    localisations = new_locs

    return localisations, timestamps


def route_localisations(route_event, time_limit, topology):
    """
    Return the localisations and timestamps of the given event route.

    :param route_event: RouteEvent
    :param time_limit: time limit of the simulation
    :param topology: Topology object corresponding to the route used
    :return: tuple of lists, localisations and timestamps
    """
    route = route_event.data["route"]
    durations = route_event.data["time"]

    current_time = route_event.timestamp

    localisations = []
    timestamps = []

    for i in range(len(route)):
        # compute current time
        current_time += durations[i]

        # we stop at simulation time limit
        if current_time > time_limit:
            break

        # append the localisation and time data
        if isinstance(route[i], tuple):
            localisations.append(route[i])
        else:
            localisations.append(topology.position_localisation(route[i]))

        timestamps.append(current_time)

    return localisations, timestamps


def add_agent_id(feature, element, agent_id=None):
    if agent_id is None:
        agent_id = element.id
    feature["properties"]["agent_id"] = agent_id


def add_icon_type(feature, element, icon_type=None):
    if icon_type is None:
        icon_type = element.icon
    feature["properties"]["icon_type"] = icon_type


def add_agent_type(feature, element, agent_type=None):
    if agent_type is None:
        agent_type = element.type
    feature["properties"]["agent_type"] = agent_type



from starling_sim.basemodel.agent.vehicles.vehicle import Vehicle
from starling_sim.basemodel.trace.events import (
    IdleEvent,
    RequestEvent,
    StopEvent,
    StaffOperationEvent,
    ServiceEvent,
)
from starling_sim.basemodel.agent.requests import Stop
from starling_sim.utils.utils import PlanningChange
from starling_sim.utils.constants import SERVICE_INIT, SERVICE_UP, SERVICE_PAUSE, SERVICE_END


from copy import copy


class ServiceVehicle(Vehicle):
    """
    This class describes a vehicle providing transport services (pickup, dropoff)
    """

    SCHEMA = {
        "properties": {
            "dwell_time": {
                "title": "Dwell time",
                "description": "Time spent when serving stops",
                "type": "integer",
                "minimum": 0,
                "default": 30,
            },
            "trip_id": {
                "advanced": True,
                "title": "Initial trip ID",
                "description": "Identifier of the vehicle's initial trip",
                "type": "string",
            },
            "operator_id": {
                "title": "Operator ID",
                "description": "ID of the operator managing this agent",
                "type": "string",
            },
            "depot": {
                "title": "Depot",
                "description": "Id of the associated depot of the service vehicle",
                "type": "string",
            },
        },
        "required": ["operator_id"],
    }

    SERVICE_INIT = SERVICE_INIT
    SERVICE_UP = SERVICE_UP
    SERVICE_PAUSE = SERVICE_PAUSE
    SERVICE_END = SERVICE_END

    def __init__(
        self,
        simulation_model,
        agent_id,
        origin,
        seats,
        operator_id,
        dwell_time=30,
        trip_id=None,
        depot=None,
        **kwargs
    ):
        super().__init__(simulation_model, agent_id, origin, seats, **kwargs)

        # service operator, managing the fleet
        self.operator = simulation_model.agentPopulation.get_agent(operator_id)

        # list of trip ids to be realised by the service vehicle (chronological order)
        self.tripList = None

        # id of the trip realised by the service vehicle, used to match user stops
        self.tripId = trip_id

        # halt duration while processing a stop
        self.dwellTime = dwell_time

        # vehicle depot
        if depot is not None:
            depot = self.operator.depotPoints[depot]
        self.depot = depot

        # planning of the service vehicle, consists in a list of Stop objects
        self.planning = []

        # boolean indicating if the vehicle is idle
        self.isIdle = False

        # event triggered to send a signal to service vehicle
        self.signalEvent_ = self.sim.scheduler.new_event_object()

        # service status
        self.serviceStatus = SERVICE_INIT

    # signaling

    def trigger_signal(self):
        former_event = self.signalEvent_

        self.signalEvent_ = self.sim.scheduler.new_event_object()

        former_event.succeed()

    # event tracing

    def trace_event(self, event):
        if isinstance(event, StopEvent):
            for request in event.pickups + event.dropoffs:
                request.agent.trace_event(event)

        if isinstance(event, StaffOperationEvent):
            if event.structure is not None:
                event.structure.trace_event(event)
            for target in event.targets:
                target.trace_event(event)

        # add event to own trace
        super().trace_event(event)

    def update_service_status(self, new_status_value: str):
        """
        Update the value of the vehicle's service status and trace a ServiceEvent.

        See possible values in starling_sim.utils.constants

        :param new_status_value: new service status value
        """
        if new_status_value not in [SERVICE_INIT, SERVICE_UP, SERVICE_PAUSE, SERVICE_END]:
            raise ValueError("Unsupported service status value: " + str(new_status_value))
        self.trace_event(
            ServiceEvent(self.sim.scheduler.now(), self.serviceStatus, new_status_value)
        )
        self.serviceStatus = new_status_value

    # stop processing

    def process_stop_(self, stop):
        # get the stops to process
        dropoff = None
        pickup = None

        if stop.type == Stop.STOP_POINT:
            dropoff = stop.dropoffList
            pickup = stop.pickupList
        elif stop.type == Stop.PUT_REQUEST:
            dropoff = stop
        elif stop.type == Stop.GET_REQUEST:
            pickup = stop
        elif stop.type is None:
            self.planning.remove(stop)
            return
        else:
            self.log_message("Unknown stop type {} to process".format(stop.type))

        # trace a stop event
        stop_event = StopEvent(self.sim.scheduler.now(), self.operator, self, self.tripId, stop)

        # compute the dwell time duration
        dwell_time = self.compute_dwell_time(stop)
        # set the effective departure time of the current stop
        if stop.type == Stop.STOP_POINT:
            stop.effectiveDepartureTime[self.tripId] = self.sim.scheduler.now() + dwell_time
        else:
            stop.effectiveDepartureTime = self.sim.scheduler.now() + dwell_time

        # process the dropoffs
        processed_dropoff = []
        if dropoff is not None:
            if isinstance(dropoff, list):
                processed_dropoff = self.process_stop_list(dropoff)
            else:
                # remove stop from planning
                self.planning.remove(stop)

                processed_dropoff = self.process_user_stop(dropoff)

            if processed_dropoff is not None and processed_dropoff != []:
                self.log_message("Dropped off {}".format(processed_dropoff))

        # trace dropoffs
        stop_event.set_dropoffs(processed_dropoff, self.sim.scheduler.now())

        yield self.execute_process(self.spend_time_(dwell_time))

        # process the pickups
        processed_pickup = []
        if pickup is not None:
            # remove stop from planning
            self.planning.remove(stop)

            if isinstance(pickup, list):
                processed_pickup = self.process_stop_list(pickup)
            else:
                processed_pickup = self.process_user_stop(pickup)

            if processed_pickup is not None and processed_pickup != []:
                self.log_message("Picked up {}".format(processed_pickup))

        # trace pickups
        stop_event.set_pickups(processed_pickup, self.sim.scheduler.now())

        self.trace_event(stop_event)

    def process_stop_list(self, stop_list):
        """
        Method called for processing the stop point pickup or dropoff list

        :param stop_list: list of user stops to be processed
        :return: list of agent ids that where processed
        """

        # list of ids of the agents that were processed
        processed = []

        # make a copy of the list so stops can be removed from the original
        list_copy = copy(stop_list)

        # process all user stops of the list
        for user_stop in list_copy:
            # process user stop and get the agent id
            result = self.process_user_stop(user_stop)

            # if an agent was indeed processed
            if result is not None:
                # remove the user stop from the list
                stop_list.remove(user_stop)

                # add the agent id to the list of processed agents
                processed.append(result)

        # return the list of processed agents
        return processed

    def process_user_stop(self, user_stop):
        """
        Process the user stop if it is relevant,
        by calling either the pickup or dropoff method.

        The stop is processed if its trip matches the service vehicle,
        and if it respects the capacity constraints of the service vehicle

        :param user_stop: UserStop object
        :return: Boolean indicating if the user stop has been processed
        """

        # test if stop should be processed by this service vehicle
        if not self.serve_stop(user_stop):
            return

        # TODO : remove when dynamic travel times are added
        # log if process time is not now
        process_time = user_stop.get_process_time()
        if process_time is not None and process_time != self.sim.scheduler.now():
            self.log_message(
                "Stop {} should be processed at {}".format(user_stop, process_time), 30
            )

        request = self.operator.requests[user_stop.requestId]
        agent_id = request.agent.id
        agent = self.sim.agentPopulation.get_agent(agent_id)

        if user_stop.type == Stop.GET_REQUEST:
            # for the pickup requests, check the service vehicle capacity
            if self.load() + request.number <= self.seats:
                self.pickup(user_stop)
                return request
            else:
                self.exceeds_capacity(user_stop)
                return None

        elif user_stop.type == Stop.PUT_REQUEST:
            # cannot dropoff user if not in vehicle
            if agent not in self.occupants:
                return None
            else:
                self.dropoff(user_stop)
                return request

        else:
            self.log_message(
                "Unknown stop type for service vehicle : {}".format(user_stop.type), 30
            )
            return None

    def pickup(self, stop):
        """
        Realise the pickup of the given stop request.

        If the request has been cancelled, remove all the
        related stops of the planning

        :param stop: UserStop object, with type GET
        """

        # get the user's request from operator
        request = self.operator.requests[stop.requestId]

        # test if request has been cancelled
        if request.success is not None and not request.success:
            self.log_message("Request {} was dismissed, cannot pickup".format(request.id))
            # remove all related stops
            for stop in self.planning:
                if stop.type == stop.GET_REQUEST and stop.requestId == request.id:
                    self.planning.remove(stop)
            return

        # update request success
        request.success = True
        request.result = self
        if stop.requestedTime is not None:
            request.waitSequence += [self.sim.scheduler.now() - stop.requestedTime]

        # trace request event
        request_event = RequestEvent(self.sim.scheduler.now(), request)
        request.agent.trace_event(request_event)

        # get service vehicle
        request.agent.get_vehicle(self)

        # wake-up user
        request.pickup_succeed()

    def dropoff(self, stop):
        """
        Realise the dropoff of the given stop request

        :param stop: UserStop object, with type PUT
        """

        # get the user's request from operator
        request = self.operator.requests[stop.requestId]

        # trace detour
        if request.directTravelTime is not None:
            detour = int(
                (request.dropoff.arrivalTime - request.pickup.departureTime)
                - request.directTravelTime
            )
            request.waitSequence += [detour]

        # leave service vehicle
        request.agent.leave_vehicle()

        # wake-up user
        request.dropoff_succeed()

        # move the request to the fulfilled collection
        # del self.operator.requests[stop.requestId]
        self.operator.fulfilled[stop.requestId] = request

    def serve_stop(self, user_stop):
        """
        Determine if the stop should be processed by the service vehicle.

        By default, use the trip id.

        :param user_stop: UserStop
        :return: boolean indicating if the stop should be processed
        """

        return user_stop.trip == self.tripId

    def exceeds_capacity(self, user_stop):
        """
        Execute the procedure for stops that cannot be picked up because of capacity.

        This method is called when a GET request cannot be served because
        of the capacity constraint.

        For instance, trigger the stop pickup event, or let the agent wait for another vehicle.

        :param user_stop: UserStop object
        """

        self.log_message("Cannot serve stop {}, the vehicle is full".format(user_stop))

    def compute_dwell_time(self, stop):
        """
        Compute the dwell time of the vehicle when processing the stop.

        The dwell time is spent between the dropoffs and the pickups.
        The computation method may differ depending if the stop is a StopPoint or
        a UserStop, or on the number of pickups and dropoffs.

        :param stop: Stop object

        :return: stop dwell time
        """

        # TODO : change, not very satisfying
        # get the dwell time for the stop
        if stop.type == Stop.STOP_POINT:
            arrival_time = stop.arrivalTime[self.tripId].pop(0)
            departure_time = stop.departureTime[self.tripId].pop(0)
            dwell_time = departure_time - arrival_time
        else:
            dwell_time = self.dwellTime

        return dwell_time

    # methods for managing move/wait when serving stops

    def go_to_next_stop_(self, next_stop, duration=None):
        """
        Move to the next stop location.

        Check changes in the tempDestination and the planning during move.

        :param next_stop: StopPoint or USerStop object
        :param duration: duration of the move
        :raises: PlanningChange if move is interrupted or next stop is changed
        """

        # get the next stop position
        self.tempDestination = next_stop.position

        # move while listening to an eventual planning change
        yield self.execute_process(self.move_(check_dest=True, duration=duration))

        # if the move was interrupted or the next stop changed, signal a planning change
        if self.position != self.tempDestination or self.planning[0] != next_stop:
            raise PlanningChange()

    # planning management and idle behaviour

    def set_planning(self, new_planning):
        """
        Set a new planning for the service vehicle.

        The planning change can imply a redirection or change the idle state
        of the service vehicle

        :param new_planning: list of Stop objects
        """

        # set the new planning
        self.planning = new_planning

        # change tempDestination, allowing dynamic redirection
        if len(self.planning) != 0:
            self.tempDestination = self.planning[0].position
        else:
            self.tempDestination = self.position

        # wake up the service vehicle if it was idle
        if self.isIdle and len(self.planning) != 0:
            try:
                self.current_process.succeed()
            except RuntimeError:
                self.log_message("Was already awake but had empty planning", 30)

    def insert_in_planning(self, request, indexes):
        """
        Insert the stops of the given request in the planning.

        :param request: TripRequest
        :param indexes: tuple of indexes for the insertion
        """

    def test_idle_(self):
        """
        Execute a certain behavior if the vehicle is idle.

        The vehicle is considered idle if its planning is empty. The idle
        behaviour is provided by the vehicle's operator.

        :return: yield the idle behaviour process. Return time being idle
        """

        # if planning is empty
        if not self.planning:
            # vehicle starts being idle
            self.isIdle = True
            start_time = self.sim.scheduler.now()
            yield self.signalEvent_ | self.execute_process(
                self.operator.idle_behaviour_(self), True
            )

            # trace idle time
            self.isIdle = False
            idle_duration = self.sim.scheduler.now() - start_time
            if idle_duration > 0:
                self.trace_event(IdleEvent(self.sim.scheduler.now(), idle_duration))

                return idle_duration

        return 0

    def return_to_depot_(self, service_status=None):
        """
        Move to service vehicle depot.

        :param service_status: optional new service status value
        """
        # get depot position
        if self.depot is None:
            self.log_message("No depot to return to")
        else:
            self.tempDestination = self.depot.position
            yield self.execute_process(self.move_())

        # optionally update service status value
        if service_status is not None:
            self.update_service_status(service_status)



class AgentPopulation:
    """
    Describes one or several collections of agents

    This class must be extended to give a concrete implementation of the storage of the
    simulation's population(s), e.g. using a dict {ID : Agent}
    """

    def __init__(self):
        """
        The constructor must be extended for the needs of the storage method
        """

        self.population = None

    def get_total_population(self):
        """
        Give access to the total population of the simulation

        :return: list of Agent objects
        """

    def get_agent(self, agent_id, agent_population=None):
        """
        Get the agent corresponding to the given id.

        The agent population may be provided, or it can be
        left to None if it is unknown.

        :param agent_id: id of the agent
        :param agent_population: population of the agent, None if unknown

        :return: agent
        """

    def new_population(self, population_name):
        """
        Return a population with the given name.

        If the population does not exist, it is created.

        :param population_name: str
        :return: population associated to the given name
        """

    def new_agent_in(self, agent, population_names):
        """
        Add the agent to the given population(s)

        :param agent: Agent object
        :param population_names: list or str corresponding to population names
        """



"""
This module contains functions for the creation of the data folder tree and
the import of example scenarios from Tellae's Google Drive.
"""

import logging
import os
import subprocess

from starling_sim.utils import paths
from starling_sim.utils.utils import create_if_not_exists
from starling_sim.utils.testing import SIMULATION_TEST_DATA_FOLDER


# data tree generation


def create_data_tree():
    """
    Create the data tree according to the paths stored in paths.py.
    """

    logging.info("Creating data tree as described in simulator_sim.utils.paths\n")

    # data folder
    logging.info("Creating folder {}".format(paths.data_folder()))
    create_if_not_exists(paths.data_folder())

    logging.info("Creating folder {}".format(paths.common_inputs_folder()))
    create_if_not_exists(paths.common_inputs_folder())

    # environment folder
    logging.info("Creating folder {}".format(paths.environment_folder()))
    create_if_not_exists(paths.environment_folder())

    # OSM graphs folder
    logging.info("Creating folder {}".format(paths.osm_graphs_folder()))
    create_if_not_exists(paths.osm_graphs_folder())

    # graph speeds folder
    logging.info("Creating folder {}".format(paths.graph_speeds_folder()))
    create_if_not_exists(paths.graph_speeds_folder())

    # GTFS feeds folder
    logging.info("Creating folder {}".format(paths.gtfs_feeds_folder()))
    create_if_not_exists(paths.gtfs_feeds_folder())

    # models folder
    logging.info("Creating folder {}".format(paths.models_folder()))
    create_if_not_exists(paths.models_folder())


# TODO : do a nicer import, without the demand files at root the reference and outputs folders,
#  and accept a list of modes to import.
def import_examples():
    """
    Import the example scenarios from the test folders.
    """

    # import the scenarios from the test data folder
    import_examples_from_test_data(SIMULATION_TEST_DATA_FOLDER)

    # if we're not in the starling folder, also get the scenarios from the starling folder
    list_dir = os.listdir(".")
    if "starling_sim" not in list_dir:
        import_examples_from_test_data(paths.starling_folder() + SIMULATION_TEST_DATA_FOLDER)


def import_examples_from_test_data(test_data_folder):
    """
    Copy the example data from the given test folder to the data folder.

    Caution : this will reproduce the folder tree of the test folder. Only the
    custom data folder will be used.

    :param test_data_folder: data folder to copy
    """

    logging.info("Copying contents of {}".format(test_data_folder))

    list_dir = os.listdir(test_data_folder)

    dest = paths.data_folder()

    for name in list_dir:
        path = test_data_folder + name

        sh_copy(path, dest)


def sh_copy(src, dst):
    """
    Copy the source file/folder in the destination folder.

    We use the recursive option of cp, so folders will be copied recursively.

    :param src: path to the source folder
    :param dst: path to the destination folder
    """

    subprocess.run(["cp", "-r", src, dst])


# # example scenarios import functions
#
# def import_example_scenario(model_code):
#     """
#     Import example scenarios of the model from the remote server.
#
#     :param model_code: code of the model to import
#     """
#
#     # check the existence of the data folder
#     if not os.path.exists(paths.data_folder()):
#         raise ModuleNotFoundError("The data folder tree must be created before importing example scenarios")
#
#     # import example scenarios of the model
#     if model_code not in example_scenarios:
#         logging.info("No example scenario is available for model {} yet".format(model_code))
#         return
#     else:
#         logging.info("Importing example scenarios for model {}".format(model_code))
#
#     # get the link of the examples archive file
#     file_id = example_scenarios[model_code]
#
#     # import and unzip the archive
#     import_folder_from_file_id(file_id, model_code, paths.models_folder())
#
#
# def import_example_environment():
#     """
#     Import the examples environment data from the remote server.
#     """
#
#     logging.info("Importing environment for example scenarios")
#
#     # import osm graphs for examples
#     import_folder_from_file_id(OSM_GRAPHS_ID, "osmGraph", paths.osm_graphs_folder(), files_only=True)
#
#     # import graph speeds for examples
#     import_folder_from_file_id(GRAPH_SPEEDS_ID, "graphSpeeds", paths.graph_speeds_folder(), files_only=True)
#
#     # import gtfs feeds for example
#     pass

# def import_folder_from_file_id(file_id, folder_name, destination_path, files_only=False):
#     """
#     Download files from the given link and delete the archive.
#
#     This method is implemented for use with dropbox links leading to folders.
#
#     :param file_id: id of the file to download
#     :param folder_name: name of the downloaded folder
#     :param destination_path: path to download destination
#     :param files_only: only import files and not the folder tree
#     """
#
#     archive_name = folder_name + ".zip"
#     archive_path = destination_path + archive_name
#
#     # build the download link from file id
#     dl_link = google_drive_download_format.format(file_id)
#
#     logging.info("Downloading files from {} in {}\n".format(dl_link, destination_path))
#
#     # download files from the link
#
#     os.system("wget --no-check-certificate -O {} '{}'".format(archive_path, dl_link))
#
#     # extract the archive. If files_only, ignore the folder tree (-j option of unzip)
#     if files_only:
#         os.system("unzip -u -j -d {} {}".format(destination_path, archive_path))
#     else:
#         os.system("unzip -u -d {} {}".format(destination_path, archive_path))
#
#     # delete the archive
#     os.system("rm {}".format(archive_path))
#
#     # add a newline
#     print()



from starling_sim.basemodel.simulation_model import SimulationModel
from starling_sim.models.SB_VS_R.input import Input
from starling_sim.models.SB_VS_R.output import Output
from starling_sim.models.SB_VS_R.operator import Operator
from starling_sim.models.SB_VS_R.repositioning_staff import RepositioningStaff
from starling_sim.models.SB_VS.user import User
from starling_sim.basemodel.agent.vehicles.station_based_vehicle import StationBasedVehicle
from starling_sim.basemodel.agent.stations.vehicle_sharing_station import VehicleSharingStation


class Model(SimulationModel):
    """Simple model for the simulation of a one-way station-based vehicle sharing,
    with repositioning operations"""

    name = "Station-based vehicle-sharing with repositioning model"

    agent_type_class = {
        "user": User,
        "vehicle": StationBasedVehicle,
        "station": VehicleSharingStation,
        "staff": RepositioningStaff,
        "operator": Operator,
    }

    leaving_codes = {"FAIL_GET": "Failed to get a vehicle"}

    modes = {
        "user": ["walk"],
        "vehicle": ["station"],
        "station": [None, "walk", "staff"],
        "staff": [None],
        "operator": {"staff": "staff", "fleet": "station"},
    }

    input_class = Input
    output_class = Output



from starling_sim.models.SB_VS.model import Model as SB_VS_model
from starling_sim.models.SB_VS_R.model import Model as SB_VS_R_model
from starling_sim.models.FF_VS.model import Model as FF_VS_model
from starling_sim.models.PT.model import Model as PT_model
from starling_sim.simulation_scenario import SimulationScenario
from starling_sim.utils.paths import model_import_path
from starling_sim.utils.utils import create_sub_scenarios

import logging
import time
import importlib

models_dict = {"SB_VS": SB_VS_model, "SB_VS_R": SB_VS_R_model, "FF_VS": FF_VS_model, "PT": PT_model}

model_codes = list(models_dict.keys())


class ModelSimulator:
    """The simulator runs a simulation model"""

    def __init__(self, model):
        """
        Initialize the simulator and its elements

        :param model: SimulationModel object describing the simulation to run
        """

        self.simulationModel = model

    def setup_simulation(self):
        """
        Prepares the simulation model for running
        """

        self.simulationModel.setup()

    def run_simulation(self):
        """
        Call the run method of the simulation model
        """

        self.simulationModel.run()

    def generate_output(self):
        """
        Generates an output of the simulation at its current state
        """

        self.simulationModel.generate_output()

    def init_simulator_from_parameters(simulation_scenario, pkg):
        """
        Returns a simulator initialised according to the model code
        contained in the given parameters

        :param simulation_scenario: SimulationParameters containing "code"
        :param pkg: name of the source package

        :return: ModelSimulator object
        """

        # get the Model class
        model_class = ModelSimulator.get_model_class(simulation_scenario["code"], pkg)

        logging.info(
            "Initialising simulation model: {} ({})\n".format(
                model_class.name, simulation_scenario.model
            )
        )

        # create a new instance of the simulation model
        try:
            simulation_model = model_class(simulation_scenario)
        except TypeError as e:
            logging.error(
                "Instantiation of {} failed with message :\n {}".format(model_class.__name__, e)
            )
            raise e

        # initialise and return the ModelSimulator object
        simulator = ModelSimulator(simulation_model)
        return simulator

    init_simulator_from_parameters = staticmethod(init_simulator_from_parameters)

    def get_model_class(model_code, pkg):
        """
        Get the Model class of the simulation model.

        :param model_code: model code
        :param pkg: name of the source package

        :return: SimulationModel subclass
        """

        # if the code corresponds to a public model, fetch the Model class in models_dict
        if model_code in models_dict:
            model_class = models_dict[model_code]

        # otherwise, import the Model class using importlib
        else:
            model_path = model_import_path(starling_pkg=pkg, model_code=model_code)

            # try to import the module corresponding to the model code
            try:
                module = importlib.import_module(model_path)
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "Cannot find the module '{}'.\n"
                    "    Maybe there is an error in the model code ? "
                    "Or maybe you forgot to use the -p option of main.py ?".format(model_path)
                )

            # try to get the Model class from the model module
            try:
                model_class = module.Model
            except AttributeError as e:
                logging.error(
                    "Cannot find the class Model in {}.models.{}.model".format(pkg, model_code)
                )
                raise e

        return model_class

    get_model_class = staticmethod(get_model_class)


def launch_simulation(scenario_path, pkg):
    """
    Realises the initialisation, setup, run and output of the simulation
    using the given parameters file. Displays log of execution times

    :param scenario_path: path to scenario folder
    :param pkg: name of the source package
    """

    if scenario_path is None:
        raise ValueError("No scenario folder provided to simulation launcher.")

    # initialise simulation scenario from folder path
    simulation_scenario = SimulationScenario(scenario_path)
    # read simulation parameters
    simulation_scenario.get_scenario_parameters()

    # init the simulator
    simulator = ModelSimulator.init_simulator_from_parameters(simulation_scenario, pkg)

    # setup the simulator
    simulator.setup_simulation()

    # run the simulation
    simulator.run_simulation()

    # generate simulation output
    simulator.generate_output()

    logging.info("End of Starling execution\n")

    logging.shutdown()



from starling_sim.basemodel.output.output_factory import OutputFactory
from starling_sim.basemodel.output.kpi_output import KpiOutput
from starling_sim.basemodel.output.kpis import (
    MoveKPI,
    SuccessKPI,
    DestinationReachedKPI,
    GetVehicleKPI,
    LeaveSimulationKPI,
)
from starling_sim.basemodel.output.information_factory import ActivityInformation


class Output(OutputFactory):
    def setup_kpi_output(self):
        # kpi output classes

        # users kpi
        move_kpi = MoveKPI()
        success_kpi = SuccessKPI(["nbFailedRequest"])
        destination_kpi = DestinationReachedKPI()
        leave_simulation_kpi = LeaveSimulationKPI()
        users_kpi_output = KpiOutput(
            "user", [move_kpi, success_kpi, destination_kpi, leave_simulation_kpi]
        )

        # vehicles kpi
        move_kpi = MoveKPI()
        get_vehicle_kpi = GetVehicleKPI()
        vehicles_kpi_output = KpiOutput("vehicle", [move_kpi, get_vehicle_kpi])

        self.kpi_outputs = [users_kpi_output, vehicles_kpi_output]

    def setup_geojson_output(self):
        super().setup_geojson_output()

        activity_information = ActivityInformation()

        information_factories = {"vehicle": [activity_information], "user": [activity_information]}

        self.geojson_output.set_information_factories(information_factories)



from starling_sim.basemodel.agent.agent import Agent
from starling_sim.basemodel.agent.requests import TripRequest, UserStop, StopPoint
from starling_sim.utils.utils import (
    geopandas_polygon_from_points,
    points_in_zone,
    json_load,
    load_schema,
)
from starling_sim.utils.paths import (
    gtfs_feeds_folder,
    scenario_agent_input_filepath,
)
from starling_sim.utils.constants import STOP_POINT_POPULATION, ADD_STOPS_COLUMNS

import pandas as pd
from typing import Union


class Operator(Agent):
    """
    Class describing an operator of a transport service.

    Operators manage a fleet of vehicle, and receive requests for the service
    they provide.

    The assignment of these requests is managed by a Dispatcher object.
    """

    OPERATION_PARAMETERS_SCHEMA = {
        "type": "object",
        "title": "Operation parameters",
        "required": [],
        "properties": {
            "routes": {
                "advanced": True,
                "type": "array",
                "items": {"type": "string"},
                "title": "GTFS routes to keep in the public transport simulation",
                "description": "List of route ids that should be present in the simulation. If null, keep all routes.",
            },
            "max_travel_time": {
                "type": ["string", "null"],
                "title": "Max travel time formula",
                "description": "Python expression used to evaluate the maximum travel time for each trip [seconds]. "
                "It can use the trip direct travel time value by using the "
                "following placeholder: {direct_travel_time}. The expression must evaluate to an "
                "object that can be cast or rounded using Python int().",
                "examples": ["{direct_travel_time} * 1.5", "{direct_travel_time} + 900", "1800"],
                "default": None,
            },
        },
    }

    SCHEMA = {
        "properties": {
            "mode": {
                "type": "object",
                "title": "Operator networks",
                "description": "Networks of the operator's agents",
                "properties": {
                    "fleet": {
                        "title": "Fleet network",
                        "description": "Default network of the operator's fleet",
                        "type": "string",
                    },
                    "staff": {
                        "title": "Staff network",
                        "description": "Default network of the operator's staff",
                        "type": "string",
                    },
                },
                "required": ["fleet"],
            },
            "fleet_dict": {
                "type": "string",
                "title": "Fleet population",
                "description": "Population of the operator's fleet agents",
            },
            "staff_dict": {
                "x-display": "hidden",
                "title": "Staff population",
                "description": "Population of the operator's staff agents",
                "type": "string",
            },
            "zone_polygon": {
                "x-display": "hidden",
                "title": "Service area file",
                "description": "Geojson input file describing the service area",
                "type": "string",
            },
            "depot_points": {
                "advanced": True,
                "title": "Depot points",
                "description": "List of depot points descriptors",
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "id": {"type": "string"},
                        "coordinates": {
                            "type": "array",
                            "items": {"type": "number"},
                            "minItems": 2,
                            "maxItems": 2,
                        },
                        "name": {"type": "string"},
                    },
                    "required": ["id", "coordinates"],
                },
                "default": [],
            },
            "extend_graph_with_stops": {
                "advanced": True,
                "title": "Extend graph with stops",
                "description": "Extend the graph with service stop points",
                "type": "boolean",
                "default": False,
            },
            "operation_parameters": OPERATION_PARAMETERS_SCHEMA,
        },
        "required": ["fleet_dict"],
        "remove_props": ["icon"],
    }

    DISPATCHERS = {}

    @classmethod
    def get_schema(cls):
        schema = cls.compute_schema()

        if len(cls.DISPATCHERS) > 0:
            enum = []

            operation_parameters_schema = schema["properties"]["operation_parameters"]

            for key in cls.DISPATCHERS.keys():
                if "title" in cls.DISPATCHERS[key]:
                    title = cls.DISPATCHERS[key]["title"]
                else:
                    title = key

                dispatchers_ops = dict()
                required_dispatchers_ops = []

                if "online" in cls.DISPATCHERS[key]:
                    online_schema = cls.DISPATCHERS[key]["online"].SCHEMA
                    if isinstance(online_schema, str):
                        online_schema = load_schema(online_schema)
                    dispatchers_ops = {**online_schema["properties"]}
                    if "required" in online_schema:
                        for prop in online_schema["required"]:
                            if prop not in required_dispatchers_ops:
                                required_dispatchers_ops.append(prop)

                if "punctual" in cls.DISPATCHERS[key]:
                    punctual_schema = cls.DISPATCHERS[key]["punctual"].SCHEMA
                    if isinstance(punctual_schema, str):
                        punctual_schema = load_schema(punctual_schema)
                    dispatchers_ops = {**punctual_schema["properties"]}
                    if "required" in punctual_schema:
                        for prop in punctual_schema["required"]:
                            if prop not in required_dispatchers_ops:
                                required_dispatchers_ops.append(prop)

                dispatcher_schema = {
                    "title": title,
                    "properties": {
                        "dispatcher": {"type": "string", "const": key, "title": "Dispatch method"},
                        **dispatchers_ops,
                    },
                    "required": required_dispatchers_ops,
                }

                enum.append(dispatcher_schema)

            operation_parameters_schema["oneOf"] = enum

        return schema

    @classmethod
    def compute_schema(cls):
        schema = super().compute_schema()

        parent_class = cls.__bases__[0]
        operation_parameters_schema = cls.OPERATION_PARAMETERS_SCHEMA
        if (
            issubclass(parent_class, Operator)
            and operation_parameters_schema != parent_class.OPERATION_PARAMETERS_SCHEMA
        ):
            if isinstance(operation_parameters_schema, str):
                operation_parameters_schema = load_schema(operation_parameters_schema)
            cls.update_class_schema(
                schema["properties"]["operation_parameters"], operation_parameters_schema, cls
            )

        return schema

    def __init__(
        self,
        simulation_model,
        agent_id,
        fleet_dict,
        mode=None,
        staff_dict=None,
        depot_points=None,
        zone_polygon=None,
        operation_parameters=None,
        extend_graph_with_stops=False,
        **kwargs
    ):
        """
        Initialise the service operator with the relevant properties.

        The operator class present many different attributes. All are not
        necessarily used when when describing a transport system operator.

        :param simulation_model: SimulationModel object
        :param agent_id: operator's id
        :param fleet_dict: name of the population that contains the operator's fleet
        :param staff_dict: name of the population that contains the operator's staff
        :param depot_points: list of dicts describing operator depot points
        :param zone_polygon: list of GPS points delimiting the service zone
        :param operation_parameters: additional parameters used for service operation
        :param extend_graph_with_stops: boolean indicating if the graph should be extended with the stop points
        :param kwargs:
        """

        Agent.__init__(self, simulation_model, agent_id, mode=mode, **kwargs)

        # data structures containing the service information

        # parent operator id
        self.parentOperatorId = None

        # parameters determining the service operation
        self.operationParameters = None
        self.init_operation_parameters(operation_parameters)

        # graph extension boolean
        self.extend_graph_with_stops = extend_graph_with_stops

        # set the fleet population (agents that provide the main transport service)
        self.fleet_name = fleet_dict
        self.fleet = self.sim.agentPopulation.new_population(fleet_dict)

        # set the staff population (agents that manage the fleet and operations)
        self.staff_dict_name = staff_dict
        self.staff = self.sim.agentPopulation.new_population(staff_dict)

        # data structure containing the main service structure, if there is one
        self.service_info = None
        self.init_service_info()

        # DataFrame containing the shapes of the service lines with columns
        # stop_id_A	stop_name_A	stop_id_B stop_name_B lon lat sequence distance distance_proportion
        self.line_shapes = None
        self.init_line_shapes()

        # GeoDataFrame representing the service zone of the operator
        self.serviceZone = None
        self.init_zone(zone_polygon)

        # a dict of the service stop points {id: StopPoint}
        self.stopPoints = dict()
        self.init_stops()

        # set the dict containing the depot points of the service
        self.depotPoints = dict()
        self.init_depot_points(depot_points)

        # trip count, used when generating trips
        self.tripCount = 0

        # a dict of service trips {trip_id: [agent, planning]}
        self.trips = dict()
        self.init_trips()

        # requests management

        # request count
        self.requestCount = 0
        # request dict
        self.requests = dict()
        # fulfilled requests
        self.fulfilled = dict()
        # left out requests
        self.leftOutRequests = []
        # minimum horizon for booking a trip
        self.bookingDeadline = 0

        # request dispatcher called punctually
        self.punctual_dispatcher = None

        # dispatcher called to handle requests online
        self.online_dispatcher = None

        self.init_dispatchers()

    # properties

    @property
    def servicePoints(self):
        """
        Return a dict mapping both stop points and depot points.
        """
        return {**self.stopPoints, **self.depotPoints}

    # attributes initialisation methods

    def init_operation_parameters(self, operation_parameters):
        """
        Initialise the operation parameters of the operator.

        If the class attribute OPERATION_PARAMETERS_SCHEMA is provided, it is
        used to validate the given parameters and set default values if needed.

        :param operation_parameters: dict of operational parameters
        """
        if operation_parameters is None:
            operation_parameters = dict()
        self.operationParameters = operation_parameters

    def init_service_info(self):
        """
        Import and initialise the data structures containing
        service information and line shapes.
        """

    def init_line_shapes(self):
        """
        Set the service lines shapes file according to the dedicated parameter.
        """

        if "line_shapes_file" in self.operationParameters:
            line_shapes_path = gtfs_feeds_folder() + self.operationParameters["line_shapes_file"]
            self.line_shapes = pd.read_csv(line_shapes_path, sep=";")
            self.line_shapes = self.line_shapes.astype({"stop_id_A": str, "stop_id_B": str})

    def init_zone(self, zone_polygon):
        """
        Set the operator service zone according to the given polygon.

        The service zone is described by a GeoDataFrame containing a
        shapely Polygon.

        :param zone_polygon: list of points describing a polygon
        """

        if isinstance(zone_polygon, str):
            filepath = scenario_agent_input_filepath(
                self.sim.scenario.scenario_folder, zone_polygon
            )
            geojson = json_load(filepath)
            service_zone = geopandas_polygon_from_points(
                geojson["features"][0]["geometry"]["coordinates"][0]
            )
        elif zone_polygon is None:
            service_zone = None
        else:
            raise TypeError(
                "The zone_polygon parameter must be either an input filename or a list of coordinates"
            )

        self.serviceZone = service_zone

    def init_depot_points(self, depot_points):
        """
        Initialise the depotPoints attribute using the given depots information.

        Depot points are StopPoint instances but are stored in a separate attribute named depotPoints.

        :param depot_points: { id, coordinates, name } information dict
        """

        if depot_points is None:
            return

        for depot_data in depot_points:
            depot_id = depot_data["id"]
            name = depot_data["depot_name"] if "depot_name" in depot_data else None
            coordinates = depot_data["coordinates"]

            # infer depot position according to service modes
            if self.mode is None:
                depot_modes = list(self.sim.environment.topologies.values())
            else:
                depot_modes = list(self.mode.values())
            position = self.sim.environment.nearest_node_in_modes(
                [coordinates[1], coordinates[0]], depot_modes
            )

            # create depot as a StopPoint object and add it to depotPoints
            self.new_stop_point(position, depot_id, name, allow_existing=False, is_depot=True)

    def init_stops(self):
        """
        Initialise the stopPoints attribute with using simulation data.
        """

    def add_stops(self, stops_table, id_prefix=""):
        """
        Add the sequence of stops to the stopPoints dict.

        The stops table is a gtfs-like DataFrame, with the columns
        "stop_id" and "stop_name". The method also uses the
        'stop_correspondence_file", which links service stops
        to topology positions.

        :param stops_table: gtfs-like DataFrame
        :param id_prefix: prefix to add to stop ids
        """

        if not set(ADD_STOPS_COLUMNS).issubset(set(stops_table.columns)):
            raise ValueError(
                "Missing columns when adding stop points. Required columns are {} and {} are provided.".format(
                    ADD_STOPS_COLUMNS, stops_table.columns
                )
            )

        # TODO : this choice of modes is arbitrary, do better
        correspondence_modes = ["walk", self.mode["fleet"]]

        # find the nearest nodes of the stops and extend the graph if asked
        self.sim.environment.add_stops_correspondence(
            stops_table, correspondence_modes, self.extend_graph_with_stops
        )

        # browse stops and add StopPoint objects to stopPoints
        for index, row in stops_table.iterrows():
            # get the stop point id, name and position from the table
            stop_point_id = id_prefix + row["stop_id"]
            stop_point_name = row["stop_name"]
            stop_position = row["nearest_node"]

            self.new_stop_point(stop_position, stop_point_id, stop_point_name)

    def new_stop_point(
        self,
        stop_point_position,
        stop_point_id,
        stop_point_name=None,
        allow_existing=True,
        is_depot=False,
    ):
        """
        Create or get existing StopPoint and add it to the operator stop points.

        :param stop_point_id: new stop id
        :param stop_point_position: new stop position
        :param stop_point_name: new stop name
        :param allow_existing: allow existence of a stop point with same id
        :param is_depot: indicates if new stop point is a depot

        :return: corresponding StopPoint object
        """
        stop_point_population = self.sim.agentPopulation.new_population(STOP_POINT_POPULATION)
        # if the stop point already exists, get it from the stop point population
        if stop_point_id in stop_point_population:
            if allow_existing:
                stop_point = stop_point_population[stop_point_id]
                # no comparison between the two positions. The stops initialisation must be well ordered
            else:
                raise ValueError("Stop point with id {} already exists !".format(stop_point_id))

        # otherwise, create a new StopPoint object and add it to the population
        else:
            if stop_point_name is None:
                stop_point = StopPoint(stop_point_position, stop_point_id)
            else:
                stop_point = StopPoint(stop_point_position, stop_point_id, stop_point_name)
            self.sim.agentPopulation.new_agent_in(stop_point, STOP_POINT_POPULATION)

        # add the stop point to the relevant operator dict
        if is_depot:
            self.depotPoints[stop_point.id] = stop_point
        else:
            self.stopPoints[stop_point.id] = stop_point

        return stop_point

    def init_trips(self):
        """
        Initialise the trips attribute using simulation data.
        """

    def add_trip(self, agent, planning, trip_id=None):
        """
        Add a new trip to the trips dict.

        Associate the trip_id to the agent that realises the trip
        and the planning that describes the trip.

        The default trip id built as <self.id>_T<self.tripCount>.

        :param agent: agent realising the trip, or None if the agent is not known yet
        :param planning: planning of the trip
        :param trip_id: id of the trip, or None

        :return: id of the trip
        """

        # set a default trip id if not provided
        if trip_id is None:
            trip_id = self.id + "_T" + str(self.tripCount)

        # set the trip's agent and planning
        self.trips[trip_id] = [agent, planning]

        # increment trip count
        self.tripCount += 1

        # return trip id
        return trip_id

    def position_in_zone(self, position):
        """
        Test if the given position belongs to the service zone.

        The method returns <self> (the operator) if the position
        is in the service zone, and None otherwise.

        :param position: topology node

        :return: either self (operator), if in zone, or None otherwise
        """

        # if no service zone, all is in service zone
        if self.serviceZone is None:
            return None

        # get position GPS localisation from global environment
        position_localisation = self.sim.environment.get_localisation(position)

        # test if localisation is in service zone
        in_zone = points_in_zone(position_localisation, self.serviceZone)

        # return self if in zone
        if in_zone:
            return self
        else:
            return None

    def init_dispatchers(self):
        """
        Initialise the punctual_dispatcher and online_dispatcher attributes.
        """

        if "dispatcher" not in self.operationParameters:
            return
        else:
            dispatcher = self.operationParameters["dispatcher"]

        if dispatcher not in self.DISPATCHERS:
            raise ValueError(
                "Unsupported operation parameter 'dispatcher' value '{}' (see schema)".format(
                    dispatcher
                )
            )

        dispatcher_classes = self.DISPATCHERS[dispatcher]

        parameters = self.dispatcher_parameters()

        try:
            if "online" in dispatcher_classes:
                self.online_dispatcher = dispatcher_classes["online"].__new__(
                    dispatcher_classes["online"]
                )
                self.online_dispatcher.__init__(**parameters)

            if "punctual" in dispatcher_classes:
                self.punctual_dispatcher = dispatcher_classes["punctual"].__new__(
                    dispatcher_classes["punctual"]
                )
                self.punctual_dispatcher.__init__(**parameters)

        except (TypeError, KeyError) as e:
            raise ValueError(
                "Instantiation of operator dispatcher failed with message :\n {}".format(str(e)), 30
            )

    def dispatcher_parameters(self):
        return {"simulation_model": self.sim, "operator": self, "verb": True}

    # new requests management

    def new_request(self, agent, number):
        """
        Create a new request object to the operator, and return it to the requesting agent.

        In the basic Operator's method, the request object is a TripRequest,
        the event is a simple SimPy event, and the request id looks like "R$self.request_count$"

        :param agent: requesting agent
        :param number: number of seats requested
        :return: Request object
        """

        # generate a new request id
        request_id = "R" + str(self.requestCount)

        # initialise a new Request object
        request = TripRequest(agent, self.sim.scheduler.now(), self, number, request_id)
        request.set_request_event()

        # add it to the request list and increment the request count
        self.requestCount += 1
        self.requests[request_id] = request

        # return the request
        return request

    def create_trip_request(
        self,
        agent,
        number,
        pickup_position,
        dropoff_position,
        pickup_request_time,
        dropoff_request_time=None,
        pickup_max_time=None,
        dropoff_max_time=None,
        pickup_stop_point=None,
        dropoff_stop_point=None,
        direct_travel_time=None,
        max_travel_time=None,
        trip_id=None,
    ):
        """
        Create a TripRequest object containing the given trip constraints.

        :param agent:
        :param number:
        :param pickup_position:
        :param dropoff_position:
        :param pickup_request_time:
        :param dropoff_request_time:
        :param pickup_max_time:
        :param dropoff_max_time:
        :param pickup_stop_point:
        :param dropoff_stop_point:
        :param direct_travel_time:
        :param max_travel_time:
        :param trip_id:
        :return:
        """

        # create the TripRequest object
        trip_request = self.new_request(agent, number)

        # create the pickup UserStop
        pickup = UserStop(
            TripRequest.GET_REQUEST,
            pickup_position,
            trip_request.id,
            requested_time=pickup_request_time,
            max_time=pickup_max_time,
            stop_point_id=pickup_stop_point,
        )

        # create the dropoff UserStop
        dropoff = UserStop(
            TripRequest.PUT_REQUEST,
            dropoff_position,
            trip_request.id,
            requested_time=dropoff_request_time,
            max_time=dropoff_max_time,
            max_travel_time=max_travel_time,
            stop_point_id=dropoff_stop_point,
        )

        # set the request user stops
        trip_request.set_stops(pickup, dropoff)

        # set direct travel time
        trip_request.directTravelTime = direct_travel_time

        # set trip id
        if trip_id is not None:
            trip_request.set_trip(trip_id)

        # set a timeout event for pickup
        if pickup_max_time is not None:
            duration_before_max_pickup_time = int(pickup_max_time - self.sim.scheduler.now())
            timeout_event = self.sim.scheduler.new_event_object() | self.sim.scheduler.timeout(
                duration_before_max_pickup_time
            )
            trip_request.pickupEvent = timeout_event

        return trip_request

    def assign_request(self, request):
        """
        Assign the given request to an agent of the fleet

        Call the dispatch algorithm in order to assign a vehicle
        to the given request, and signal the schedule change to
        the chosen vehicle.

        :param request:
        """

        if self.online_dispatcher is None:
            self.leftOutRequests.append(request)
        else:
            self.online_dispatcher.online_dispatch(request)

    def cancel_request(self, request):
        """
        Remove the pending queues of the given assigned request.

        For instance, remove the UserStop objects from the stop points lists.

        Called when a request is canceled by the requesting user or the operator.

        :param request: cancelled request
        """

        print("hello")

    def build_trip_request(
        self,
        number,
        agent=None,
        origin_position=None,
        origin_stop=None,
        origin_time=None,
        destination_position=None,
        destination_stop=None,
        destination_time=None,
        **kwargs
    ):
        """
        Build a request for a trip with the operator service from the given information.

        Depending on the operator, all parameters may not be necessary, and others mays be added .

        :param agent: agent making the trip request
        :param number: number of seats requested
        :param origin_position: trip origin position in environment
        :param origin_stop: trip origin stop id
        :param origin_time: trip origin time
        :param destination_position: trip destination position in environment
        :param destination_stop: trip destination stop id
        :param destination_time: trip destination time
        :param kwargs: other eventual parameters

        :return: TripRequest completed with trip information set
        """

    # idle service vehicles management

    def idle_behaviour_(self, agent):
        """
        Perform an idle behaviour.

        This method is called when an agent of the
        operator's fleet is idle. The agent then executes
        the behaviour until it is interrupted. The default
        behaviour is to wait indefinitely.

        :param agent: Agent object
        :return: yield a process describing an idle behaviour
        """

        yield agent.execute_process(agent.spend_time_())

    # creation of new service vehicles
    def new_service_vehicle(self, feature):
        """
        Create and initialise a new service vehicle using the model's dynamic input

        :param feature: dictionary containing the service vehicle information
        :return: newly generated ServiceVehicle
        """

        if (
            "operator_id" not in feature["properties"]
            or feature["properties"]["operator_id"] != self.id
        ):
            self.log_message(
                "Error in the 'operator' field for the generation of "
                "a service vehicle : {}".format(feature),
                30,
            )
            return

        new_service_vehicle = self.sim.dynamicInput.new_agent_input(feature)

        return new_service_vehicle

    # utils

    def evaluate_time_window(self, base_time: int, time_window: int):
        """
        Evaluate a time window from the reference and the time window width.

        If the time window width is negative, the reference time is the upper bound,
        and the lower bound is evaluated.

        :param base_time: base for the evaluation of the time window
        :param time_window: time window width

        :return: (time window_end, time_window_end) tuple
        """

        if time_window == 0:
            self.log_message("Time window cannot be equal to 0", 40)
            raise ValueError("Time window cannot be equal to 0")
        elif time_window > 0:
            start = base_time
            end = base_time + time_window
        else:
            start = base_time + time_window
            end = base_time

        return start, end

    def compute_max_travel_time(self, direct_travel_time: int) -> Union[int, None]:
        """
        Evaluate the max travel time formula to get a maximum travel time value [seconds].

        The formula is a Python expression, evaluated using the builtin eval function.
        It can use the direct travel time value by using the following placeholder: {direct_travel_time}
        It must return something that can be cast or rounded using int().

        Examples (without the brackets):
            - "{direct_travel_time} * 1.5"
            - "{direct_travel_time} + 900"
            - "1800"

        :param direct_travel_time: value of the direct travel time (int)

        :return: value of the maximum travel time (int) or None if no constraint is applied
        :raises: ValueError if the evaluation of the formula fails
        """

        # get formula from operation parameters
        max_travel_time_formula = self.operationParameters["max_travel_time"]
        if max_travel_time_formula is None:
            return None

        # format formula with direct travel time
        formula = max_travel_time_formula.format(direct_travel_time=direct_travel_time)

        try:
            # evaluate formula with eval() builtin
            max_travel_time = eval(formula)

            # cast to integer if not None
            if max_travel_time is not None:
                max_travel_time = int(max_travel_time)
        except Exception:
            msg = "Failed to evaluate the following formula for max travel time: {}".format(
                max_travel_time_formula
            )
            self.log_message(msg, 40)
            raise ValueError(msg)

        return max_travel_time



import numpy as np
from abc import ABC


class Dispatcher(ABC):
    """
    This class describes a dispatcher for a transport service.

    The dispatcher is called to affect requests online or punctually.

    The Dispatcher class allows a complete specification (punctual and online)
    and separation (from the operator and the algorithms it uses) of the dispatching strategy.

    The parameters needed to initialise and run the dispatcher should be provided
    in the operator operationParameters attribute.
    """

    #: json schema for the operation parameters
    SCHEMA = {"type": "object", "properties": {}}

    def __init__(self, simulation_model, operator, verb=False):
        """
        Initialise the dispatcher and its algorithm.

        The operator's dispatcher is initialised only once. It is then called on
        various situations to realise a dispatch.

        :param simulation_model: SimulationModel
        :param operator: id of the operator of the dispatched service
        :param verb: verbose boolean
        """

        #: simulation object
        self.sim = simulation_model

        #: transport operator id
        self.operator = operator

        #: operation parameters
        self.operationParameters = None
        if operator is not None:
            self.operationParameters = self.operator.operationParameters

        #: verbose option
        self.verb = verb

        #: dispatch algorithm
        self.algorithm = None
        self.init_algorithm()

        #: algorithm result
        self.result = None

        #: online request, in case of online dispatch
        self.online_request = None

    @property
    def name(self):
        return self.__class__.__name__

    def init_algorithm(self):
        """
        Initialise the algorithm's structure used by the dispatcher.
        """

    def dispatch(self):
        """
        Realise the complete dispatch process.
        """

        # setup the algorithm and parameters
        self.setup_dispatch()

        # run the algorithm and store the result in punctual_result
        self.run_algorithm()

        # update the plannings according to the proposed solution
        self.update_from_solution()

    def setup_dispatch(self):
        """
        Setup the dispatch with current simulation data.
        """

    def run_algorithm(self):
        """
        Run the algorithm on the current problem instance.
        """

    def update_from_solution(self):
        """
        Update the operator plannings based on the dispatch solution.
        """

    # dispatching loop for punctual algorithms

    def dispatching_loop_(self):
        self.log_message("Dispatching loop not implemented", lvl=40)
        raise NotImplementedError

    # method for online dispatch of requests

    def online_dispatch(self, request):
        self.log_message("Online dispatch method not implemented", lvl=40)
        raise NotImplementedError

    # logs and trace

    def log_message(self, message, lvl=15):
        """
        Log a message using the operator's dedicated method.

        :param message:
        :param lvl:
        """

        if self.verb or lvl >= 30:
            # add the name of the dispatcher as a prefix
            message = "{} : {}".format(self.name, message)

            # log message using the operator method
            self.operator.log_message(message, lvl=lvl)

    # utils

    def compute_matrix_of(self, stop_points, dimension, mode):
        """
        Compute the distance matrix (in the given dimension) between the given stop points,
        on the given topology graph.

        :param stop_points: list of stop point ids
        :param dimension: dimension of the distance to compute
        :param mode: travel mode between modes

        :return: matrix[i, j] = distance from stop i to stop j.
        """

        matrix = np.zeros((len(stop_points), len(stop_points)))

        for i in range(len(stop_points)):
            origin = self.operator.servicePoints[stop_points[i]].position
            _, lengths = self.sim.environment.topologies[mode].compute_dijkstra_path(
                origin, None, dimension
            )

            for j in range(len(stop_points)):
                matrix[i, j] = int(lengths[self.operator.servicePoints[stop_points[j]].position])

        if dimension == "time":
            matrix = matrix.astype(int)

        return matrix



from starling_sim.basemodel.trace.events import *
from starling_sim.basemodel.agent.requests import UserStop, StopPoint, StationRequest
from starling_sim.utils.constants import (
    PUBLIC_TRANSPORT_TYPE,
    SERVICE_INIT,
    SERVICE_UP,
    SERVICE_PAUSE,
    SERVICE_END,
)


class KPI:
    """
    Generic structure of a KPI class

    Its sub-classes compute and update specific indicator
    from given events
    """

    #: **agentId**: id of the agent
    KEY_ID = "agentId"

    def __init__(self):
        """
        The indicator_dict associates values to the indicators names

        The keys attribute correspond to the keys of the indicator_dict
        """
        self.indicator_dict = None
        self.keys = []

        self.new_indicator_dict()

    def setup(self, simulation_model):
        """
        Setup method called during simulation setup.

        :param simulation_model:
        :return:
        """
        pass

    def new_indicator_dict(self):
        """
        Reset the indicator_dict, when computing the indicator
        for a new target
        :return: None, resets directly the indicator_dict attribute
        """
        pass

    def update(self, event, agent):
        """
        Update the kpi values according to the event content and the agent.

        :param event: processed event
        :param agent: subject of the event
        :return:
        """

        if isinstance(event, InputEvent):
            self.indicator_dict[self.KEY_ID] = agent.id


class MoveKPI(KPI):
    """
    This KPI evaluates the distance and spent time for each one of the simulation modes
    """

    #: **{mode}Distance**: distance travelled in <mode> [meters]
    SUFFIX_KEY_DISTANCE = "{mode}Distance"
    #: **{mode}Time**: time travelled in <mode> [seconds]
    SUFFIX_KEY_TIME = "{mode}Time"

    def __init__(self):
        self.modes = []

        # init of indicator dict
        super().__init__()

    def setup(self, simulation_model):
        self.modes = list(simulation_model.environment.topologies.keys())
        self.new_indicator_dict()

    def new_indicator_dict(self):
        """
        Initialize the time and distance values at 0
        for the considered modes
        :return:
        """
        base_dict = {}

        for mode in self.modes:
            key = self.SUFFIX_KEY_DISTANCE.format(mode=mode)
            base_dict[key] = 0
            self.keys += [key]

            key = self.SUFFIX_KEY_TIME.format(mode=mode)
            base_dict[key] = 0
            self.keys += [key]

        self.indicator_dict = base_dict

    def update(self, event, agent):
        """
        Add travelled distances and durations
        :param agent:
        :param event:
        :return:
        """

        super().update(event, agent)

        if isinstance(event, MoveEvent):
            self.indicator_dict[self.SUFFIX_KEY_DISTANCE.format(mode=event.mode)] += event.distance
            self.indicator_dict[self.SUFFIX_KEY_TIME.format(mode=event.mode)] += event.duration


class WaitKPI(KPI):
    """
    This KPI evaluates the time spent waiting
    """

    #: **waitTime**: total traced wait time [seconds]
    KEY_WAIT = "waitTime"

    def __init__(self):
        super().__init__()
        self.keys = [self.KEY_WAIT]

    def new_indicator_dict(self):
        self.indicator_dict = {self.KEY_WAIT: 0}

    def update(self, event, agent):
        """
        Add total wait duration of the request
        :param agent:
        :param event:
        :return:
        """

        super().update(event, agent)

        if isinstance(event, RequestEvent):
            self.indicator_dict[self.KEY_WAIT] += sum(event.request.waitSequence)

        if isinstance(event, WaitEvent):
            self.indicator_dict[self.KEY_WAIT] += event.waiting_time


class OdtWaitsKPI(KPI):
    """
    This KPI evaluates the lateness in ODT requests.
    """

    #: **odtPickupWait**: series of wait times at ODT pickups [seconds]
    KEY_PICKUP_WAIT = "odtPickupWait"
    #: **odtDetour**: series of ODT detour times [seconds]
    KEY_DETOUR = "odtDetour"
    #: **odtDirectTrip**: series of ODT direct trip times [seconds]
    KEY_DIRECT_TRIP = "odtDirectTrip"

    def __init__(self):
        super().__init__()
        self.keys = [self.KEY_PICKUP_WAIT, self.KEY_DETOUR, self.KEY_DIRECT_TRIP]

    def new_indicator_dict(self):
        self.indicator_dict = {
            self.KEY_PICKUP_WAIT: "",
            self.KEY_DETOUR: "",
            self.KEY_DIRECT_TRIP: "",
        }

    def update(self, event, agent):
        """
        Add wait durations of ODT requests to KPIs.

        :param event:
        :param agent:
        :return:
        """

        # TODO : find a better condition
        if isinstance(event, StopEvent) and event.serviceVehicle.type != PUBLIC_TRANSPORT_TYPE:
            dropoff_agents = [request.agent.id for request in event.dropoffs]
            pickup_agents = [request.agent.id for request in event.pickups]

            if agent.id in dropoff_agents:
                request = event.dropoffs[dropoff_agents.index(agent.id)]
                if len(request.waitSequence) > 1:
                    if self.indicator_dict[self.KEY_DETOUR] != "":
                        self.indicator_dict[self.KEY_DETOUR] += "-"
                        self.indicator_dict[self.KEY_DIRECT_TRIP] += "-"
                    self.indicator_dict[self.KEY_DETOUR] += str(request.waitSequence[1])
                self.indicator_dict[self.KEY_DIRECT_TRIP] += str(request.directTravelTime)

            elif agent.id in pickup_agents:
                request = event.pickups[pickup_agents.index(agent.id)]
                if len(request.waitSequence) > 0:
                    if self.indicator_dict[self.KEY_PICKUP_WAIT] != "":
                        self.indicator_dict[self.KEY_PICKUP_WAIT] += "-"
                    self.indicator_dict[self.KEY_PICKUP_WAIT] += str(request.waitSequence[0])


class GetVehicleKPI(KPI):
    """
    This KPI evaluates the number of vehicle uses
    """

    #: **nbGetVehicle**: number of uses of the vehicle
    KEY_GET_VEHICLE = "nbGetVehicle"

    def __init__(self):
        super().__init__()
        self.keys = [self.KEY_GET_VEHICLE]

    def new_indicator_dict(self):
        self.indicator_dict = {self.KEY_GET_VEHICLE: 0}

    def update(self, event, agent):
        """
        Add a new use for each GetVehicleEvent
        :param agent:
        :param event:
        :return:
        """

        super().update(event, agent)

        if isinstance(event, GetVehicleEvent):
            self.indicator_dict[self.KEY_GET_VEHICLE] += event.agent.number


class SuccessKPI(KPI):
    """
    This KPI evaluates the number of failed/successful requests
    """

    #: **nbFailedGet**: number of failed get requests
    KEY_FAILED_GET = "nbFailedGet"
    #: **nbSuccessGet**: number successful get requests
    KEY_SUCCESS_GET = "nbSuccessGet"
    #: **nbFailedPut**: number of failed put requests
    KEY_FAILED_PUT = "nbFailedPut"
    #: **nbSuccessPut**: number of successful put requests
    KEY_SUCCESS_PUT = "nbSuccessPut"
    #: **nbFailedRequest**: number of failed requests
    KEY_FAILED_REQUEST = "nbFailedRequest"
    #: **nbSuccessRequest**: number of successful requests
    KEY_SUCCESS_REQUEST = "nbSuccessRequest"

    def __init__(self, indicator_selection):
        super().__init__()

        self.keys = indicator_selection

    def new_indicator_dict(self):
        base_dict = {
            self.KEY_FAILED_GET: 0,
            self.KEY_SUCCESS_GET: 0,
            self.KEY_FAILED_PUT: 0,
            self.KEY_SUCCESS_PUT: 0,
            self.KEY_FAILED_REQUEST: 0,
            self.KEY_SUCCESS_REQUEST: 0,
        }

        self.indicator_dict = base_dict

    def update(self, event, agent):
        """
        Add request events according to their success
        :param agent:
        :param event:
        :return:
        """

        super().update(event, agent)

        if isinstance(event, RequestEvent):
            if event.request.success:
                self.indicator_dict[self.KEY_SUCCESS_REQUEST] += 1
                if event.request.type == StationRequest.GET_REQUEST:
                    self.indicator_dict[self.KEY_SUCCESS_GET] += 1
                else:
                    self.indicator_dict[self.KEY_SUCCESS_PUT] += 1
            else:
                self.indicator_dict[self.KEY_FAILED_REQUEST] += 1
                if event.request.type == StationRequest.GET_REQUEST:
                    self.indicator_dict[self.KEY_FAILED_GET] += 1
                else:
                    self.indicator_dict[self.KEY_FAILED_PUT] += 1


class StaffOperationKPI(KPI):
    """
    This KPI evaluates the number of staff operations
    """

    #: **nbFailedGetStaff**: number of failed gets by staff
    KEY_FAILED_GET_STAFF = "nbFailedGetStaff"
    #: **nbSuccessGetStaff**: number of successful gets by staff
    KEY_SUCCESS_GET_STAFF = "nbSuccessGetStaff"
    #: **nbFailedPutStaff**: number of failed puts by staff
    KEY_FAILED_PUT_STAFF = "nbFailedPutStaff"
    #: **nbSuccessPutStaff**: number of successful puts by staff
    KEY_SUCCESS_PUT_STAFF = "nbSuccessPutStaff"

    def __init__(self):
        super().__init__()
        self.keys = [
            self.KEY_FAILED_GET_STAFF,
            self.KEY_SUCCESS_GET_STAFF,
            self.KEY_FAILED_PUT_STAFF,
            self.KEY_SUCCESS_PUT_STAFF,
        ]

    def new_indicator_dict(self):
        self.indicator_dict = {
            self.KEY_FAILED_GET_STAFF: 0,
            self.KEY_SUCCESS_GET_STAFF: 0,
            self.KEY_FAILED_PUT_STAFF: 0,
            self.KEY_SUCCESS_PUT_STAFF: 0,
        }

    def update(self, event, agent):
        """
        Add operations to the total
        :param agent:
        :param event:
        :return:
        """

        super().update(event, agent)

        if isinstance(event, StaffOperationEvent):
            goal = event.goal
            total = event.total
            if goal < 0:
                self.indicator_dict[self.KEY_SUCCESS_GET_STAFF] += abs(total)
                self.indicator_dict[self.KEY_FAILED_GET_STAFF] += total - goal
            elif goal > 0:
                self.indicator_dict[self.KEY_SUCCESS_PUT_STAFF] += total
                self.indicator_dict[self.KEY_FAILED_PUT_STAFF] += goal - total


class OccupationKPI(KPI):
    """
    This KPI evaluates the empty and full time and distance
    and the stock relative time/distance
    """

    def __init__(self):
        #: **emptyTime**: time spent empty [seconds]
        self.KEY_EMPTY_TIME = "emptyTime"
        #: **emptyDistance**: distance travelled empty [meters]
        self.KEY_EMPTY_DISTANCE = "emptyDistance"
        #: **fullTime**: time spent full [seconds]
        self.KEY_FULL_TIME = "fullTime"
        #: **fullDistance**: distance travelled full [meters]
        self.KEY_FULL_DISTANCE = "fullDistance"
        #: **stockTime**: stock relative time (stock*time) [seconds]
        self.KEY_STOCK_TIME = "stockTime"
        #: **stockDistance**: stock relative distance (stock*distance) [meters]
        self.KEY_STOCK_DISTANCE = "stockDistance"
        #: **maxStock**: maximum stock
        self.KEY_MAX_STOCK = "maxStock"

        super().__init__()

        self.keys = [
            self.KEY_EMPTY_TIME,
            self.KEY_EMPTY_DISTANCE,
            self.KEY_FULL_TIME,
            self.KEY_FULL_DISTANCE,
            self.KEY_STOCK_TIME,
            self.KEY_STOCK_DISTANCE,
            self.KEY_MAX_STOCK,
        ]

        self.capacity = None
        self.currentStock = None
        self.previousTime = 0
        self.currentDistance = None

    def new_indicator_dict(self):
        """
        Initialize the time and distance counts to 0.
        """

        self.indicator_dict = {
            self.KEY_EMPTY_TIME: 0,
            self.KEY_EMPTY_DISTANCE: 0,
            self.KEY_FULL_TIME: 0,
            self.KEY_FULL_DISTANCE: 0,
            self.KEY_STOCK_TIME: 0,
            self.KEY_STOCK_DISTANCE: 0,
            self.KEY_MAX_STOCK: 0,
        }
        self.capacity = None
        self.currentStock = None
        self.previousTime = 0
        self.currentDistance = None

    def get_capacity(self, element):
        """
        Get the capacity of the agent, according to its type.

        :param element:

        :return: agent's capacity
        """
        return self.capacity

    def get_initial_stock(self, element):
        """
        Get the initial stock of the agent, according to its type.

        :param element:

        :return: agent's initial stock
        """
        return self.currentStock

    def add_to_stock(self, value, timestamp):
        """
        Update the full and empty time and distance counts, according to the previous
        stock value, then updates the stock and time.

        :param value: stock change (negative for stock loss)
        :param timestamp: timestamp of the stock change event
        """

        # compute time spent with last stock
        duration = int(timestamp - self.previousTime)

        # add time to relevant time count
        if self.currentStock == 0:
            self.indicator_dict[self.KEY_EMPTY_TIME] += duration
            if self.currentDistance is not None:
                self.indicator_dict[self.KEY_EMPTY_DISTANCE] += self.currentDistance
        elif self.currentStock == self.capacity:
            self.indicator_dict[self.KEY_FULL_TIME] += duration
            if self.currentDistance is not None:
                self.indicator_dict[self.KEY_FULL_DISTANCE] += self.currentDistance

        # add stock relative time and distance
        self.indicator_dict[self.KEY_STOCK_TIME] += duration * self.currentStock
        if self.currentDistance is not None:
            self.indicator_dict[self.KEY_STOCK_DISTANCE] += self.currentDistance * self.currentStock

        # update stock and current time
        self.currentStock += value
        self.previousTime = timestamp

        if self.currentStock > self.indicator_dict[self.KEY_MAX_STOCK]:
            self.indicator_dict[self.KEY_MAX_STOCK] = self.currentStock

        # reset distance count
        if self.currentDistance is not None:
            self.currentDistance = 0

    def update(self, event, agent):
        """
        Update the stock and time counts from traced events

        :param agent:
        :param event:
        """

        super().update(event, agent)

        if isinstance(event, InputEvent):
            self.capacity = self.get_capacity(event.element)
            self.currentStock = self.get_initial_stock(event.element)
            self.indicator_dict[self.KEY_MAX_STOCK] = self.currentStock

        if isinstance(event, LeaveSimulationEvent):
            self.add_to_stock(0, event.timestamp)


class StationOccupationKPI(OccupationKPI):
    """
    This KPI evaluates the time spent in the empty and full states (of a station),
    and the stock relative time spent in the station
    """

    def __init__(self):
        super().__init__()

        self.keys = [self.KEY_EMPTY_TIME, self.KEY_FULL_TIME, self.KEY_STOCK_TIME]

    def get_capacity(self, element):
        return element.capacity

    def get_initial_stock(self, element):
        return element.initial_stock

    def update(self, event, agent):
        """
        Update the stock and time counts from request events

        :param agent:
        :param event:
        """

        super().update(event, agent)

        if isinstance(event, RequestEvent) and event.request.success:
            request = event.request

            # update time counts and current time
            if request.type == StationRequest.GET_REQUEST:
                self.add_to_stock(-1, request.timestamp)
            elif request.type == StationRequest.PUT_REQUEST:
                self.add_to_stock(1, request.timestamp)

        if isinstance(event, StaffOperationEvent):
            if event.total != 0:
                self.add_to_stock(event.total, event.timestamp)


class VehicleOccupationKPI(OccupationKPI):
    """
    This KPI evaluates the time and distance in the empty and full states (of vehicle),
    and a passenger relative distance and time.
    """

    def __init__(self):
        super().__init__()
        self.currentDistance = 0

    def new_indicator_dict(self):
        super().new_indicator_dict()
        self.currentDistance = 0

    def get_capacity(self, element):
        return element.seats

    def get_initial_stock(self, element):
        # for now, initial stock is always 0 in our simulation
        return 0

    def update(self, event, agent):
        """
        Update the stock and time/distance counts from
        get/leave vehicle and move events

        :param agent:
        :param event:
        """

        super().update(event, agent)

        if isinstance(event, GetVehicleEvent):
            self.add_to_stock(event.agent.number, event.timestamp)

        elif isinstance(event, LeaveVehicleEvent):
            self.add_to_stock(-event.agent.number, event.timestamp)

        if isinstance(event, MoveEvent):
            self.currentDistance += event.distance


class ChargeKPI(KPI):
    """
    This KPI evaluates the trips's boards and un-boards
    """

    #: **routeId**: gtfs route id
    KEY_ROUTE_ID = "routeId"
    #: **tripId**: gtfs trip id
    KEY_TRIP_ID = "tripId"
    #: **tripDirection**: gtfs trip direction
    KEY_TRIP_DIRECTION = "tripDirection"
    #: **time**: simulation timestamp of board/un-board
    KEY_TIME = "time"
    #: **stopId**: stop id of board/un-board
    KEY_STOP_ID = "stopId"
    #: **boardType**: (+1) for boards, (-1) for un-boards
    KEY_BOARD_TYPE = "boardType"
    #: **value**: numeric value of the charge change
    KEY_VALUE = "value"

    def __init__(self, non_empty_only=True, public_transport=True):
        super().__init__()

        # boolean indicating if only non empty pickups and dropoffs should be traced
        self.non_empty_only = non_empty_only

        # boolean indicating if the simulated system is public transports (with gtfs tables)
        self.public_transport = public_transport

        self.trips = None
        self.routes = None

        self.keys = [
            self.KEY_TRIP_ID,
            self.KEY_TIME,
            self.KEY_STOP_ID,
            self.KEY_BOARD_TYPE,
            self.KEY_VALUE,
        ]

        if self.public_transport:
            self.keys = [self.KEY_ROUTE_ID, self.KEY_TRIP_DIRECTION] + self.keys

        self.new_indicator_dict()

    def setup(self, simulation_model):
        if self.public_transport:
            self.trips = simulation_model.gtfs.trips
            self.routes = simulation_model.gtfs.routes

    def new_indicator_dict(self):
        self.indicator_dict = dict()

        for key in [self.KEY_ID] + self.keys:
            self.indicator_dict[key] = []

    def update(self, event, agent):
        """
        Add stop information to the list

        :param agent:
        :param event:
        :return:
        """

        if isinstance(event, StopEvent):
            if event.dropoffs:
                self.update_stop_information(event, agent)
                self.indicator_dict[self.KEY_TIME].append(event.dropoff_time)
                self.indicator_dict[self.KEY_BOARD_TYPE].append(-1)
                self.indicator_dict[self.KEY_VALUE].append(len(event.dropoffs))

            if event.pickups:
                self.update_stop_information(event, agent)
                self.indicator_dict[self.KEY_TIME].append(event.pickup_time)
                self.indicator_dict[self.KEY_BOARD_TYPE].append(1)
                self.indicator_dict[self.KEY_VALUE].append(len(event.pickups))

            if not event.pickups and not event.dropoffs and not self.non_empty_only:
                self.update_stop_information(event, agent)
                self.indicator_dict[self.KEY_TIME].append(event.timestamp)
                self.indicator_dict[self.KEY_BOARD_TYPE].append(0)
                self.indicator_dict[self.KEY_VALUE].append(0)

    def update_stop_information(self, event, agent):
        """
        Update the indicator with the information common to dropoffs and pickups.

        :param event:
        :param agent:
        """

        self.indicator_dict[self.KEY_ID].append(agent.id)

        trip_id = event.trip

        self.indicator_dict[self.KEY_TRIP_ID].append(trip_id)

        self.indicator_dict[self.KEY_STOP_ID].append(get_stop_id_of_event(event))

        if self.public_transport:
            self.indicator_dict[self.KEY_ROUTE_ID].append(
                get_route_id_of_trip(self.trips, trip_id, event)
            )
            self.indicator_dict[self.KEY_TRIP_DIRECTION].append(
                get_direction_of_trip(self.trips, trip_id)
            )


class ServiceKPI(KPI):
    """
    This KPI describes the service time of a vehicle.
    """

    #: **serviceDuration**: vehicle service duration [seconds]
    KEY_SERVICE_DURATION = "serviceDuration"

    def __init__(self):
        super().__init__()
        self.currentServiceStart = None
        self.totalServiceDuration = None
        self.keys = [self.KEY_SERVICE_DURATION]

    def new_indicator_dict(self):
        self.indicator_dict = {self.KEY_SERVICE_DURATION: "NA"}

    def update(self, event, agent):
        if isinstance(event, ServiceEvent):
            if event.new == SERVICE_UP:
                self.start_service(event.timestamp)
            elif event.new in [SERVICE_PAUSE, SERVICE_END]:
                self.close_service(event.timestamp)
            elif event.new == SERVICE_INIT:
                raise ValueError("Service state shouldn't change to {} status".format(SERVICE_INIT))
            else:
                raise ValueError("Unsupported service status " + event.new)

        if isinstance(event, LeaveSimulationEvent):
            if self.totalServiceDuration is None or self.currentServiceStart is not None:
                self.indicator_dict[self.KEY_SERVICE_DURATION] = "NA"
            else:
                self.indicator_dict[self.KEY_SERVICE_DURATION] = self.totalServiceDuration
            self.currentServiceStart = None

    def start_service(self, timestamp):
        if self.currentServiceStart is not None:
            raise ValueError("Service starts but has not ended")
        else:
            self.currentServiceStart = timestamp

    def close_service(self, timestamp):
        if self.currentServiceStart is None:
            raise ValueError("Service ends but has not started")
        else:
            duration = timestamp - self.currentServiceStart
            if self.totalServiceDuration is None:
                self.totalServiceDuration = duration
            else:
                self.totalServiceDuration += duration
            self.currentServiceStart = None


class TransferKPI(KPI):
    """
    This KPI lists the transfers realised by a user,
    with additional information such as walk distance and duration,
    wait duration, from/to trip/stop
    """

    #: **walkDistance**: walk distance of transfer [meters]
    KEY_WALK_DIST = "walkDistance"
    #: **walkTime**: walk time of transfer [seconds]
    KEY_WALK_DURATION = "walkDuration"
    #: **waitTime**: wait time of transfer [seconds]
    KEY_WAIT_TIME = "waitTime"
    #: **fromRoute**: origin route of transfer
    KEY_FROM_ROUTE = "fromRoute"
    #: **fromTrip**: origin trip of transfer
    KEY_FROM_TRIP = "fromTrip"
    #: **fromStop**: origin stop point of transfer
    KEY_FROM_STOP = "fromStop"
    #: **toRoute**: destination route of transfer
    KEY_TO_ROUTE = "toRoute"
    #: **toTrip**: destination trip of transfer
    KEY_TO_TRIP = "toTrip"
    #: **toStop**: destination stop of transfer
    KEY_TO_STOP = "toStop"

    def __init__(self):
        super().__init__()

        self.trips = None
        self.routes = None

        self.keys = [
            self.KEY_WALK_DIST,
            self.KEY_WALK_DURATION,
            self.KEY_WAIT_TIME,
            self.KEY_FROM_ROUTE,
            self.KEY_FROM_TRIP,
            self.KEY_FROM_STOP,
            self.KEY_TO_ROUTE,
            self.KEY_TO_TRIP,
            self.KEY_TO_STOP,
        ]

        # transfer variables
        self.current_walk_distance = 0
        self.current_walk_duration = 0
        self.current_wait_time = 0
        self.from_route = None
        self.from_trip = None
        self.from_stop = None
        self.to_route = None
        self.to_trip = None
        self.to_stop = None

    def setup(self, simulation_model):
        if simulation_model.gtfs is not None:
            self.trips = simulation_model.gtfs.trips
            self.routes = simulation_model.gtfs.routes

    def new_indicator_dict(self):
        self.indicator_dict = {
            self.KEY_ID: [],
            self.KEY_WALK_DIST: [],
            self.KEY_WALK_DURATION: [],
            self.KEY_WAIT_TIME: [],
            self.KEY_FROM_ROUTE: [],
            self.KEY_FROM_TRIP: [],
            self.KEY_FROM_STOP: [],
            self.KEY_TO_ROUTE: [],
            self.KEY_TO_TRIP: [],
            self.KEY_TO_STOP: [],
        }

    def update(self, event, agent):
        if isinstance(event, InputEvent):
            self.reset_variables()

        if isinstance(event, WaitEvent):
            self.current_wait_time += event.waiting_time

        elif isinstance(event, MoveEvent) and event.mode == "walk":
            self.current_walk_distance += event.distance
            self.current_walk_duration += event.duration

        elif isinstance(event, RequestEvent):
            self.current_wait_time += sum(event.request.waitSequence)

        elif isinstance(event, StopEvent):
            if isinstance(event.stop, StopPoint):
                stop_id = event.stop.id
            elif isinstance(event.stop, UserStop):
                stop_id = event.stop.stopPoint
            else:
                stop_id = None

            dropoff_agents = [request.agent.id for request in event.dropoffs]
            pickup_agents = [request.agent.id for request in event.pickups]

            if agent.id in dropoff_agents:
                self.from_trip = event.trip
                self.from_stop = stop_id

            elif agent.id in pickup_agents:
                self.to_trip = event.trip
                self.to_stop = stop_id

                self.write_variables(agent)
                self.reset_variables()

        elif isinstance(event, DestinationReachedEvent):
            self.write_variables(agent)
            self.reset_variables()

    def reset_variables(self):
        self.current_walk_distance = 0
        self.current_walk_duration = 0
        self.current_wait_time = 0
        self.from_trip = None
        self.from_stop = None
        self.to_trip = None
        self.to_stop = None

    def write_variables(self, agent):
        self.indicator_dict[self.KEY_ID].append(agent.id)
        self.indicator_dict[self.KEY_WALK_DIST].append(self.current_walk_distance)
        self.indicator_dict[self.KEY_WALK_DURATION].append(self.current_walk_duration)
        self.indicator_dict[self.KEY_WAIT_TIME].append(self.current_wait_time)
        self.indicator_dict[self.KEY_FROM_ROUTE].append(
            get_route_short_name_of_trip(self.trips, self.routes, self.from_trip)
        )
        self.indicator_dict[self.KEY_FROM_TRIP].append(self.from_trip)
        self.indicator_dict[self.KEY_FROM_STOP].append(self.from_stop)
        self.indicator_dict[self.KEY_TO_ROUTE].append(
            get_route_short_name_of_trip(self.trips, self.routes, self.to_trip)
        )
        self.indicator_dict[self.KEY_TO_TRIP].append(self.to_trip)
        self.indicator_dict[self.KEY_TO_STOP].append(self.to_stop)


# TODO : remove ? useless with TransferKPI
# class JourneyKPI(KPI):
#     """
#     This KPI evaluates the sequence of lines that compose a user journey
#     """
#
#     #: **journeySequence**: sequence of trip ids of the journey
#     KEY_JOURNEY_SEQUENCE = "journeySequence"
#
#     def __init__(self):
#
#         super().__init__()
#
#         self.keys = [self.KEY_JOURNEY_SEQUENCE]
#
#         self.trips_table = None
#         self.routes_table = None
#
#     def setup(self, simulation_model):
#
#         operator = simulation_model.agentPopulation["operators"]["OPR"]
#
#         feed = operator.service_info
#
#         self.trips_table = feed.trips
#         self.routes_table = feed.routes
#
#     def new_indicator_dict(self):
#
#         self.indicator_dict = {self.KEY_JOURNEY_SEQUENCE: ""}
#
#     def update(self, event, agent):
#         """
#         Add a new route for each PickupEvent
#         :param agent:
#         :param event:
#         :return:
#         """
#
#         super().update(event, agent)
#
#         if isinstance(event, PickupEvent):
#
#             route_id = self.trips_table.loc[self.trips_table["trip_id"] == event.trip, "route_id"].iloc[0]
#
#             route_name = self.routes_table.loc[self.routes_table["route_id"] == route_id, "route_short_name"].iloc[0]
#
#             self.indicator_dict[self.KEY_JOURNEY_SEQUENCE] += route_name + "-"
#
#         if isinstance(event, DestinationReachedEvent):
#
#             self.indicator_dict[self.KEY_JOURNEY_SEQUENCE] = self.indicator_dict[self.KEY_JOURNEY_SEQUENCE][:-1]
#


class DestinationReachedKPI(KPI):
    """
    This KPI evaluates the destination reach time
    """

    #: **destinationReachedTime**: time when destination is reached, "NA" otherwise [seconds or NA]
    KEY_DESTINATION_REACHED = "destinationReachedTime"

    def __init__(self):
        super().__init__()

        self.keys = [self.KEY_DESTINATION_REACHED]

    def new_indicator_dict(self):
        self.indicator_dict = {self.KEY_DESTINATION_REACHED: "NA"}

    def update(self, event, agent):
        """
        Add total wait duration of the request
        :param agent:
        :param event:
        :return:
        """

        super().update(event, agent)

        if isinstance(event, DestinationReachedEvent):
            self.indicator_dict[self.KEY_DESTINATION_REACHED] = event.timestamp


class LeaveSimulationKPI(KPI):
    """
    This KPI evaluates the cause of the simulation leave.
    """

    #: **leaveSimulation**: code used when leaving the simulation (see model doc)
    KEY_LEAVE_SIMULATION = "leaveSimulation"

    def __init__(self):
        super().__init__()

        self.keys = [self.KEY_LEAVE_SIMULATION]

    def new_indicator_dict(self):
        self.indicator_dict = {self.KEY_LEAVE_SIMULATION: None}

    def update(self, event, agent):
        """
        Add the cause of the LeaveSimulationEvent.

        :param event:
        :param agent:
        """

        super().update(event, agent)

        if isinstance(event, LeaveSimulationEvent):
            self.indicator_dict[self.KEY_LEAVE_SIMULATION] = event.cause


def get_route_id_of_trip(trips, trip_id, event):
    if trips is None:
        return event.serviceVehicle.operator

    trip_table = trips.loc[trips["trip_id"] == trip_id, "route_id"]

    # if trips is not in the gtfs (on-demand trips for instance)
    # try to get operator id
    if trip_table.empty:
        return event.serviceVehicle.operator

    return trip_table.iloc[0]


def get_direction_of_trip(trips, trip_id):
    if trips is None:
        return ""

    trip_table = trips.loc[trips["trip_id"] == trip_id, "direction_id"]

    # ignore trips that are not in the gtfs (on-demand trips for instance)
    if trip_table.empty:
        return ""

    return trip_table.iloc[0]


def get_route_short_name_of_trip(trips, routes, trip_id):
    if trips is None or routes is None or trip_id is None:
        return None

    trip_table = trips.loc[trips["trip_id"] == trip_id, "route_id"]

    # ignore trips that are not in the gtfs (on-demand trips for instance)
    if trip_table.empty:
        return ""

    route_id = trip_table.iloc[0]
    route_short_name = routes.loc[routes["route_id"] == route_id, "route_short_name"].iloc[0]

    return route_short_name


def get_stop_id_of_event(event):
    stop_id = None

    if isinstance(event.stop, StopPoint):
        stop_id = event.stop.id
    elif isinstance(event.stop, UserStop):
        stop_id = event.stop.stopPoint

    if stop_id is None:
        stop_id = ""

    return stop_id



from starling_sim.basemodel.agent.operators.operator import Operator
from starling_sim.utils.utils import get_sec, new_point_feature, stop_table_from_gtfs
from starling_sim.utils.constants import PUBLIC_TRANSPORT_TYPE
import pandas as pd
import numpy as np
import sys


class PublicTransportOperator(Operator):
    """
    Class describing an operator of a public transport service
    """

    SCHEMA = {
        "remove_props": ["fleet_dict"],
        "properties": {
            "extend_graph_with_stops": {
                "advanced": True,
                "title": "Extend graph with stops",
                "description": "Indicate if the transport network should be extended with stop points",
                "type": "boolean",
                "default": True,
            }
        },
    }

    OPERATION_PARAMETERS_SCHEMA = {
        "properties": {
            "use_shortest_path": {
                "title": "Use shortest path between stops",
                "description": "Follow the network shortest path between stops."
                " Otherwise, the trajectory is a straight line",
                "type": "boolean",
                "default": False,
            },
            "nb_seats": {
                "type": "object",
                "properties": {
                    "0": {
                        "title": "Tramways capacity",
                        "description": "Number of seats of tramways",
                        "type": "integer",
                        "minimum": 0,
                        "default": 440,
                    },
                    "1": {
                        "title": "Subways capacity",
                        "description": "Number of seats of subways",
                        "type": "integer",
                        "minimum": 0,
                        "default": 722,
                    },
                    "2": {
                        "title": "Trains capacity",
                        "description": "Number of seats of trains",
                        "type": "integer",
                        "minimum": 0,
                        "default": 2600,
                    },
                    "3": {
                        "title": "Buses capacity",
                        "description": "Number of seats of buses",
                        "type": "integer",
                        "minimum": 0,
                        "default": 166,
                    },
                },
                "required": ["0", "1", "2", "3"],
            },
            "service_vehicle_prefix": {
                "title": "Service vehicle prefix",
                "description": "Prefix of the service vehicles ids, prepended to the trip id",
                "type": "string",
                "minLength": 1,
                "default": "V-",
            },
            "line_shapes_file": {
                "type": "string",
                "title": "Line shapes file",
                "description": "Name of the file containing shapes of public transport lines (stored with GTFS feeds)",
                "pattern": "(.)*(.csv)",
            },
        },
        "required": ["use_shortest_path", "nb_seats", "service_vehicle_prefix"],
    }

    ROUTE_TYPE_ICONS = {"0": "tram", "1": "subway", "2": "train", "3": "bus"}

    FLEET_TYPE = PUBLIC_TRANSPORT_TYPE

    def __init__(self, simulation_model, agent_id, fleet_dict, **kwargs):
        super().__init__(simulation_model, agent_id, fleet_dict, **kwargs)

    def init_service_info(self):
        # get the complete timetables from simulation model
        feed = self.sim.gtfs

        # filter feed using simulation parameters

        # filter gtfs date
        date = self.sim.scenario["date"]
        feed = feed.restrict_to_dates([date])

        # filter gtfs routes
        if "routes" in self.operationParameters:
            feed = feed.restrict_to_routes(self.operationParameters["routes"])

        if feed.stop_times.empty:
            self.log_message(
                "Filtered gtfs is empty. Are you sure that there "
                "are trips running on the given date ?",
                30,
            )

        # keep all stops from the original gtfs
        feed.stops = self.sim.gtfs.stops

        self.service_info = feed

    def init_stops(self):
        # add the stop points of active trips of the gtfs
        stops_table = stop_table_from_gtfs(self.service_info, active_stops_only=True)
        self.add_stops(stops_table)

    def init_trips(self):
        # add the active trips of the gtfs
        stop_times = self.service_info.get_stop_times()

        # only keep the first arrival time and filter with simulation time limit
        min_stop_sequence = stop_times["stop_sequence"].min()
        stop_times = stop_times[stop_times["stop_sequence"] == min_stop_sequence]
        stop_times = stop_times.sort_values(by="arrival_time")
        stop_times["arrival_time_num"] = stop_times["arrival_time"].apply(get_sec)
        stop_times = stop_times[stop_times["arrival_time_num"] < self.sim.scenario["limit"]]

        # create the planning of each trip
        for index, row in stop_times.iterrows():
            trip_planning = self.build_planning_of_trip(row["trip_id"])
            self.add_trip(None, trip_planning, trip_id=row["trip_id"])

    def get_route_and_direction_of_trip(self, trip_id):
        """
        Evaluate the route and direction of a trip using the service info.

        :param trip_id:
        :return:
        """

        trips = self.service_info.get_trips()

        route_id = trips.loc[trips["trip_id"] == trip_id, "route_id"].values[0]
        direction_id = trips.loc[trips["trip_id"] == trip_id, "direction_id"].values[0]

        return route_id, direction_id

    def create_public_transport_fleet(self):
        """
        Create a fleet for the public transport service.

        Services vehicles are generated using the block_id field of the gtfs.
        When no block id is provided for a trip, a vehicle dedicated to this single
        trip is generated.
        """

        # get trips ordered by first arrival time
        trips = self.service_info.get_trips()
        stop_times = self.service_info.get_stop_times()
        min_stop_sequence = stop_times["stop_sequence"].min()
        stop_times = stop_times[stop_times["stop_sequence"] == min_stop_sequence]
        stop_times["arrival_time_num"] = stop_times["arrival_time"].apply(get_sec)
        stop_times = stop_times[stop_times["arrival_time_num"] < self.sim.scenario["limit"]]
        trips = pd.merge(trips, stop_times, on="trip_id")
        trips = trips.sort_values(by="arrival_time")

        # get the block_id values
        if "block_id" not in trips.columns:
            trips["block_id"] = np.nan
            block_ids = [np.nan]
        else:
            block_ids = trips.drop_duplicates(subset="block_id")["block_id"].values

        for block_id in block_ids:
            if pd.isna(block_id):
                block_trips = trips[pd.isna(trips["block_id"])]["trip_id"].values
                # if block_id is nan, generate separate vehicles for each trip
                for trip_id in block_trips:
                    vehicle_id = self.operationParameters["service_vehicle_prefix"] + trip_id
                    self.create_service_vehicle(vehicle_id, [trip_id])
            else:
                block_trips = trips[trips["block_id"] == block_id]["trip_id"].values
                # otherwise, generate a vehicle with multiple with multiple trips
                vehicle_id = self.operationParameters["service_vehicle_prefix"] + block_id
                self.create_service_vehicle(vehicle_id, block_trips)

    def create_service_vehicle(self, agent_id, trip_id_list):
        """
        Create a new service vehicle in the simulation.

        :param agent_id: id of the new vehicle
        :param trip_id_list: list of trips to be realised by service the vehicle
        """

        # get the route type from the first trip of the vehicle
        trips = self.service_info.get_trips()
        trip_id = trip_id_list[0]
        route_id = trips.loc[trips["trip_id"] == trip_id, "route_id"].iloc[0]
        routes = self.service_info.get_routes()
        route_type = str(routes.loc[routes["route_id"] == route_id, "route_type"].iloc[0])

        # get the agent icon from route type
        if route_type in self.ROUTE_TYPE_ICONS:
            agent_icon = self.ROUTE_TYPE_ICONS[route_type]
        else:
            self.log_message("Route type {} not in icon dict".format(route_type), 30)
            agent_icon = "bus"

        # get the number of seats from route type
        nb_seats_dict = self.operationParameters["nb_seats"]
        if route_type in nb_seats_dict:
            nb_seats = nb_seats_dict[route_type]
        else:
            self.log_message("Route type {} not in nb_seats dict".format(route_type), 30)
            nb_seats = sys.maxsize

        # get the origin from the first planning
        origin = self.trips[trip_id][1][0].position

        # build the service vehicle properties dict
        input_properties = {
            "agent_id": agent_id,
            "origin": origin,
            "operator_id": self.id,
            "seats": nb_seats,
            "agent_type": self.FLEET_TYPE,
            "mode": self.mode["fleet"],
            "icon": agent_icon,
        }

        # build a geojson feature
        feature = new_point_feature(properties=input_properties)

        # generate the service vehicle in the simulation
        new_vehicle = self.new_service_vehicle(feature)

        # set the new vehicle trip list
        new_vehicle.tripList = trip_id_list

        for trip_id in trip_id_list:
            self.trips[trip_id][0] = agent_id

    def build_planning_of_trip(self, trip_id):
        """
        Build the planning corresponding to the trip from the gtfs.

        :param trip_id: gtfs trip_id
        :return: trip planning, list of StopPoint objects
        """

        trip_planning = []

        # get the trip stop times
        stop_times = self.service_info.get_stop_times()
        stop_times = stop_times[stop_times["trip_id"] == trip_id]

        for index, row in stop_times.iterrows():
            stop_point = self.stopPoints[row["stop_id"]]

            # append stop point to planning
            trip_planning += [stop_point]

            # set service times in stop point
            if trip_id not in stop_point.arrivalTime:
                stop_point.arrivalTime[trip_id] = [get_sec(row["arrival_time"])]
            else:
                stop_point.arrivalTime[trip_id] += [get_sec(row["arrival_time"])]

            if trip_id not in stop_point.departureTime:
                stop_point.departureTime[trip_id] = [get_sec(row["departure_time"])]
            else:
                stop_point.departureTime[trip_id] += [get_sec(row["departure_time"])]

        return trip_planning

    def loop_(self):
        self.create_public_transport_fleet()
        yield self.execute_process(self.spend_time_(1))



from starling_sim.basemodel.trace.events import *
from starling_sim.basemodel.agent.vehicles.vehicle import Vehicle
from starling_sim.basemodel.agent.stations.vehicle_sharing_station import VehicleSharingStation
from starling_sim.utils.constants import END_OF_SIM_LEAVE

from abc import ABC
import logging


class InformationFactory(ABC):
    """
    Generic structure of a geojson information factory.

    Its subclasses put together specific information to add in the geojson features,
    in the 'information' property.
    """

    #: Default key used in the 'information' property
    DEFAULT_KEY = "key"

    def __init__(self, information_key=None):
        """
        Initialise the structures containing the information to add to each feature.

        Information is stored based on an event structure. An initial value is stored,
        and then only changes in the information value are listed, along with the
        timestamp of the change.

        :param information_key:
        """

        # simulation model
        self.sim = None

        # list of values for the information
        self.values = []

        # list of timestamps for the information
        self.timestamps = []

        # name of the information, used as a key in the information property dict
        if information_key is None:
            self.key = self.DEFAULT_KEY
        else:
            self.key = information_key

    def setup(self, simulation_model):
        """
        Setup method called during simulation setup.

        :param simulation_model:
        """
        self.sim = simulation_model

    def get_dict(self):
        """
        Get the dict to add in the 'information' property.

        :return: dict with keys 'values' and 'timestamps'.
        """

        # log a warning if values and timestamps are of different lengths
        if len(self.values) != len(self.timestamps):
            logging.warning(
                "Values and timestamps of information {} have different lengths.".format(self.key)
            )
            return dict()

        information_dict = {"values": self.values, "timestamps": self.timestamps}

        # reset the information structures
        self.new_information_dict()

        return information_dict

    def new_information_dict(self):
        """
        Reset the structures containing the information.
        """
        self.values = []
        self.timestamps = []

    def append_value_and_timestamp(self, value, timestamp):
        """
        Append a value and its timestamp to the information structures.

        Avoids appending only one of the structures.

        :param value:
        :param timestamp:
        """

        self.values.append(value)
        self.timestamps.append(timestamp)

    def update(self, event, agent):
        """
        Update the information structures according to the event content and agent.

        :param event:
        :param agent:
        """
        pass


class ActivityInformation(InformationFactory):
    """
    This factory traces the activity of agents.
    """

    DEFAULT_KEY = "activity"

    def update(self, event, agent):
        if isinstance(event, InputEvent):
            self.append_value_and_timestamp(0, 0)

            start_of_activity_timestamps = agent.trace.eventList[1].timestamp
            self.append_value_and_timestamp(1, start_of_activity_timestamps)

        if isinstance(event, LeaveSimulationEvent):
            end_of_activity_timestamps = agent.trace.eventList[-2].timestamp

            self.append_value_and_timestamp(0, end_of_activity_timestamps)


class StockInformation(InformationFactory):
    """
    This factory traces the stock of vehicles and stations.
    """

    DEFAULT_KEY = "stock"

    def update(self, event, agent):
        if isinstance(agent, VehicleSharingStation):
            if isinstance(event, InputEvent):
                self.append_value_and_timestamp(agent.initial_stock, 0)

            elif isinstance(event, RequestEvent) and event.request.success:
                request = event.request

                if request.type == request.GET_REQUEST:
                    value = self.values[-1] - 1
                elif request.type == request.PUT_REQUEST:
                    value = self.values[-1] + 1
                else:
                    return

                self.append_value_and_timestamp(value, event.timestamp)

        elif isinstance(agent, Vehicle):
            if isinstance(event, InputEvent):
                self.append_value_and_timestamp(0, 0)

            elif isinstance(event, GetVehicleEvent):
                value = self.values[-1] + 1
                self.append_value_and_timestamp(value, event.timestamp)

            elif isinstance(event, LeaveVehicleEvent):
                value = self.values[-1] - 1
                self.append_value_and_timestamp(value, event.timestamp)


class DelayInformation(InformationFactory):
    """
    This factory traces the delay of service vehicles compared to the timetable.
    """

    DEFAULT_KEY = "delay"

    def update(self, event, agent):
        if isinstance(event, InputEvent):
            self.append_value_and_timestamp(0, event.timestamp)

        elif isinstance(event, StopEvent):
            # get the theoretical departure time
            service_info = agent.operator.service_info
            stop_times = service_info.get_stop_times()
            theo_departure_time = stop_times[
                (stop_times["trip_id"] == event.trip) & (stop_times["stop_id"] == event.stop.id)
            ]["departure_time_num"].values[0]

            # compute the delay
            delay = int((event.pickup_time - theo_departure_time) / float(60))

            # format the delay information
            if delay > 0:
                delay = "+" + str(delay)
            else:
                delay = str(delay)

            # append the information
            self.append_value_and_timestamp(delay, event.pickup_time)



import logging

from starling_sim.basemodel.population.agent_population import AgentPopulation


class DictPopulation(AgentPopulation):
    """
    Stores the agent populations of the simulation in Python dictionaries
    Theses dicts are stored in a global dict
    """

    def __init__(self, population_names):
        """
        Initialises the dicts with the given names
        :param population_names: list of names for the populations
        """

        super().__init__()

        self.population = {}
        self.base_names = population_names

        for name in population_names:
            self.new_population(name)

    def get_total_population(self):
        agents = []

        for name in self.base_names:
            agents = agents + list(self.population[name].values())

        return agents

    def get_agent(self, agent_id, agent_population=None):
        if agent_population is not None:
            pop_dict = self.population[agent_population]
            if agent_id in pop_dict:
                return pop_dict[agent_id]

        else:
            for pop_dict in self.population.values():
                if agent_id in pop_dict:
                    return pop_dict[agent_id]

    def new_population(self, population_name):
        if population_name in self.population:
            return self.population[population_name]
        else:
            new_population = dict()
            self.population[population_name] = new_population
            return new_population

    def new_agent_in(self, agent, population_names):
        if isinstance(population_names, str):
            population_names = [population_names]

        for population_name in population_names:
            # raise KeyError exception if agent already exists (and then cancel agent add)
            if agent.id in self.population[population_name]:
                logging.error(
                    "Agent {} already exists in population {}, cancelled introduction of duplicate".format(
                        agent.id, population_name
                    )
                )
                raise KeyError

            self.population[population_name][agent.id] = agent

    def __getitem__(self, item):
        """
        Method called when using agentPopulation[item]
        :param item: name of the population accessed
        :return: self.population[item]
        """

        if item not in self.population:
            logging.warning("Trying to get unknown population " + item)
            return None
        return self.population[item]

    # useless ?
    def __setitem__(self, key, value):
        """
        Method called when using agentPopulation[key] = value
        :param key: name of the population accessed
        :param value: population dict
        :return:
        """

        if key in self.population:
            logging.warning("Assigning dict to an existing population " + key)
        self.population[key] = value



"""
Free-floating vehicle-sharing user
"""

from starling_sim.basemodel.agent.persons.person import Person
from starling_sim.basemodel.agent.requests import Request


class User(Person):
    """
    This class describes a free-floating vehicle-sharing user
    """

    def __init__(self, simulation_model, agent_id, origin, destination, **kwargs):
        super().__init__(simulation_model, agent_id, origin, destination, **kwargs)

        # all for now

    def loop_(self):
        """
        Main loop of a free-floating vehicle-sharing user
        """

        # loop on trying to get closest vehicle
        yield self.execute_process(self.request_loop_(self.maxTries))

        if self.vehicle is None:
            # if failed to get a vehicle, leave the system
            self.leave_simulation("FAIL_GET")

        # get closest point to destination on the adequate topology and move to it
        vehicle_mode = self.vehicle.mode
        self.tempDestination = self.closest_walkable_node_of(
            vehicle_mode, position=self.destination
        )
        yield self.execute_process(self.move_shortest_with_vehicle_())

        # leave vehicle
        self.leave_vehicle()

        # end trip
        yield self.execute_process(self.walk_to_destination_())

        return

    def try_new_request_(self):
        """
        Try to get the closest vehicle available.

        :return: Request object, completed according to request result
        """

        # create a new request
        vehicle_request = Request(self, self.sim.scheduler.now(), request_type=Request.GET_REQUEST)

        # get relevant vehicles
        vehicles = []

        for vehicle in self.sim.agentPopulation["vehicle"].values():
            # only check vehicles with occupants == []
            if not vehicle.occupants:
                vehicles.append(vehicle)

        # compute closest vehicle
        closest_vehicle, route = self.sim.environment.closest_object(
            self.position, vehicles, True, "walk", return_path=True, n=3
        )

        # if no vehicle available, fail
        if closest_vehicle is None:
            vehicle_request.success = False
            return vehicle_request

        # walk to vehicle position
        self.tempDestination = closest_vehicle.position
        yield self.execute_process(self.move_(route=route))

        # if vehicle is available, take it
        if closest_vehicle.position == self.position and not closest_vehicle.occupants:
            self.get_vehicle(closest_vehicle)
            vehicle_request.success = True
            vehicle_request.result = closest_vehicle
        else:
            vehicle_request.success = False

        return vehicle_request



from starling_sim.utils.simulation_logging import ALGO_LOGGER, ALGO_LEVEL


class Algorithm:
    """
    This class describes a generic structure for the algorithms of the simulation
    """

    NAME = "unnamed algorithm"

    def __init__(self, simulation_model=None, operator=None, verb=False):
        # simulation objects
        self.sim = simulation_model
        self.operator = operator

        # logging options
        self.verbose = verb

        # result attribute
        self.result = None

    def run(self):
        """
        Core method of the Algorithm class, called to run the algorithm.

        This method must be implemented for all algorithms.

        Algorithm inputs are accessed in the class attributes.

        :return: None, the algorithm result is stored in self.result
        """

        self.log_message("Start running the algorithm {}".format(self.NAME))

    def log_message(self, message, lvl=ALGO_LEVEL):
        """
        Logs the message in the logger using the class data

        :param message: message displayed in log
        :param lvl: level value, default is ALGO_LEVEL
        """

        if self.verbose or lvl >= 30:
            extra_params = {"alg_name": self.NAME}

            ALGO_LOGGER.log(lvl, message, extra=extra_params)

    def __str__(self):
        """
        Give a string display to the algorithm.

        :return: string with algorithm information
        """

        return "Algorithm ({})".format(self.NAME)



"""
This module contains a config dict with simulation related values.

Config values can be provided to the simulator using a json file name after the _CONFIG_FILE constant.

For a complete description of the config specification and its default values,
see the JSON schema (schemas/starling_config.schema.json):

.. literalinclude:: /../starling_sim/schemas/starling_config.schema.json
    :language: json
"""

from starling_sim.utils.utils import json_load, add_defaults_and_validate

import os

#: name of the config file
_CONFIG_FILE = "starling_config.json"

#: filename of the config schema
_CONFIG_SCHEMA = "starling_config.schema.json"


def _init_config():
    """
    Initialise config variables from a config file or from the default schema values.
    """

    # if a config file exists at the project root, use it to define config values
    if os.path.exists(_CONFIG_FILE):
        user_config = json_load(_CONFIG_FILE)
    else:
        user_config = {}

    # use the default values for the keys that are not provided
    return add_defaults_and_validate(user_config, _CONFIG_SCHEMA)


#: config dict used in Starling
config = _init_config()



"""
This module contains constants and default values for the simulation.
"""

# keys and names used in the simulation

#: name of the population containing the simulation stop points
STOP_POINT_POPULATION = "stop_point"

#: agent_type of the agents generated by PublicTransportOperator
PUBLIC_TRANSPORT_TYPE = "public_transport"

#: required DataFrame columns of the table describing the stop points when calling add_stops()
ADD_STOPS_COLUMNS = ["stop_id", "stop_name", "stop_lat", "stop_lon"]

# base leaving codes

#: default leaving code, when the loop terminates without LeavingSimulation exception
DEFAULT_LEAVE = "DEFAULT"

#: leaving code signaling a completed loop
SUCCESS_LEAVE = "SUCCESS"

#: leaving code signaling the end of the simulation
END_OF_SIM_LEAVE = "END_OF_SIM"

#: leaving code signaling a simulation error
SIM_ERROR_LEAVE = "SIM_ERROR"

#: dict of base leaving codes and their descriptions
BASE_LEAVING_CODES = {
    DEFAULT_LEAVE: "No explicit leave call",
    SUCCESS_LEAVE: "Successful loop",
    END_OF_SIM_LEAVE: "End of simulation",
    SIM_ERROR_LEAVE: "Simulation error",
}

# service statuses

#: Status used when service has not started yet
SERVICE_INIT = "INIT"
#: Status used when service started and is up
SERVICE_UP = "UP"
#: Status used when service has started but is paused
SERVICE_PAUSE = "PAUSE"
#: Status used when service has ended
SERVICE_END = "END"

# filename formats

#: filename of the run summary file
RUN_SUMMARY_FILENAME = "run_summary.json"

#: format of the ALIENS operators usage diagram
OPERATORS_USAGE_FORMAT = "{prefix}_operators.png"

#: format of the ALIENS objective function graph
OBJECTIVE_GRAPH_FORMAT = "{prefix}_objective.png"



from starling_sim.basemodel.input.dynamic_input import DynamicInput


class Input(DynamicInput):
    def pre_process_input_dict(self, input_dict):
        # TODO : use a default fleet and stations name when there is no operator

        if input_dict["agent_type"] == "station":
            operator = self.sim.agentPopulation.population["operator"][input_dict["operator_id"]]

            # add station population
            input_dict["population"] = operator.stations_dict_name

        elif input_dict["agent_type"] == "user":
            if "origin_station" in input_dict:
                user_station = self.sim.agentPopulation["station"][input_dict["origin_station"]]
                input_dict["origin"] = user_station.position

            if "destination_station" in input_dict:
                user_station = self.sim.agentPopulation["station"][
                    input_dict["destination_station"]
                ]
                input_dict["destination"] = user_station.position

        elif input_dict["agent_type"] == "staff":
            # add origin at depot
            # TODO : adapt for multiple depots
            operator = self.sim.agentPopulation.population["operator"][input_dict["operator_id"]]
            input_dict["origin"] = list(operator.depotPoints.values())[0].position



import argparse
import logging
import os
import json

from starling_sim.model_simulator import launch_simulation, ModelSimulator
from starling_sim.utils.data_tree import create_data_tree, import_examples
from starling_sim.utils.simulation_logging import DEFAULT_LOGGER_LEVEL, setup_logging
from starling_sim.utils import paths
from starling_sim import __version__


def run_main():
    # create a command line parser

    parser = argparse.ArgumentParser(description="Starling agent-based simulation framework")

    parser.add_argument("scenario_path", help="path to scenario folder.", nargs="?")

    parser.add_argument(
        "-l",
        "--level",
        help="specify the logger level. See simulation_logging.py for more information.",
        type=int,
        default=None,
    )

    parser.add_argument(
        "-D",
        "--data-tree",
        help="generate the data tree according to the paths stored in paths.py and exit.",
        action="store_true",
    )

    parser.add_argument(
        "-e",
        "--examples",
        help="import the example scenarios of the given model codes from the test folder "
        "and exit. Generate the data tree folders if they don't exist.",
        action="store_true",
    )

    parser.add_argument(
        "-S",
        "--sphinx",
        action="store_true",
        help="generate the project documentation using Sphinx and exit.",
    )

    parser.add_argument(
        "-p",
        "--package",
        help="indicate an alternative name for the base package of starling",
        type=str,
        action="store",
        default="starling_sim",
    )

    parser.add_argument("--data-folder", help="Change the path to the data folder", default=None)

    parser.add_argument(
        "-J",
        "--json-schema",
        help="generate json schemas for each agent type of the model",
        metavar="MODEL_CODE",
        type=str,
        action="store",
    )

    parser.add_argument("-v", "--version", action="version", version=__version__)

    # parse the command line

    input_args = parser.parse_args()

    # setup logging
    if input_args.level is None:
        input_args.level = DEFAULT_LOGGER_LEVEL
    setup_logging(input_args.level)

    if input_args.json_schema is not None:
        model_class = ModelSimulator.get_model_class(input_args.json_schema, input_args.package)
        schemas = model_class.get_agent_type_schemas()
        print(json.dumps(schemas, indent=4))
        exit(0)

    # documentation generation
    if input_args.sphinx:
        os.system("./docs/sphinx-doc.sh")
        exit(0)

    if input_args.data_folder is not None:
        if input_args.data_folder.endswith("/"):
            paths._DATA_FOLDER = input_args.data_folder
        else:
            paths._DATA_FOLDER = input_args.data_folder + "/"

    # data tree create
    if input_args.data_tree:
        create_data_tree()
        exit(0)

    # example scenarios
    if input_args.examples:
        # make sure the data tree is setup
        create_data_tree()

        # import the example scenarios and environment
        import_examples()
        exit(0)

    # launch simulation
    logging.info("Launching Starling {}\n".format(__version__))
    launch_simulation(input_args.scenario_path, input_args.package)



"""
Starling setup script.

See license in LICENSE.txt.
"""

import setuptools
import os
from starling_sim import __version__

# short description of the project
DESC = "Agent-based framework for mobility simulation"

# get long description from README.md
# with open("README.md", "r") as fh:
#     LONG_DESC = fh.read()
LONG_DESC = r"""
**Starling** is an agent-based framework for mobility simulation implemented in Python. 
The framework provides users with tools to quickly develop agent-based simulators of a specific transport service.
Among these tools, there is a number of generic transport classes that can be used in several transport contexts,
accessible input and output formats (JSON, GeoJSON) with JSON schema specifications, visualisation and KPI outputs, etc.

**Starling** also comes with three simulation models :

 - A free-floating vehicle-sharing service
 - A station-based vehicle-sharing service
 - A station-based vehicle-sharing service with repositioning

**For now, Starling is not adapted to be used as a package : users should rather clone the project 
locally to develop and run new simulations.**

Read the `docs`_ or see the project on `Github`_.

Citation info: Leblond, V, Desbureaux, L, Bielecki, V., 2020. `A new agent-based software for designing and 
optimizing emerging mobility services : application to city of Rennes`_. 
*Proceedings of the European Transport Conference*, 11 September 2020, Milan.

.. _Github: https://github.com/tellae/starling
.. _docs: https://starling.readthedocs.io/en/latest/
.. _A new agent-based software for designing and optimizing emerging mobility services \: application to city of Rennes : https://aetransport.org/past-etc-papers/conference-papers-2020?abstractId=6706&state=b
"""

# list of classifiers from the PyPI classifiers trove
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Scientific/Engineering :: GIS",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "License :: CeCILL-B Free Software License Agreement (CECILL-B)",
]

# only specify install_requires if not in RTD environment
if os.getenv("READTHEDOCS") == "True":
    INSTALL_REQUIRES = []
else:
    with open("requirements.txt") as f:
        INSTALL_REQUIRES = [line.strip() for line in f.readlines()]

# call setup
setuptools.setup(
    name="starling-sim",
    version=__version__,
    license="CECILL-B",
    author="Tellae",
    author_email="starling@tellae.fr",
    description=DESC,
    long_description_content_type="text/x-rst",
    long_description=LONG_DESC,
    url="https://github.com/tellae/starling",
    packages=setuptools.find_packages() + ["starling_sim/schemas"],
    classifiers=CLASSIFIERS,
    python_requires=">=3.6",
    install_requires=INSTALL_REQUIRES,
    include_package_data=True,
)
